from copy import copy

import h5py
import numpy as np
import pytest
from numpy.testing import assert_equal

from threedigrid_builder.interface import GridAdminOut


@pytest.fixture(scope="session")
def h5_out(tmpdir_factory, grid_all):
    path = tmpdir_factory.mktemp("h5") / "gridadmin.h5"

    grid_all.meta.has_0d = True

    with GridAdminOut(path) as out:
        out.write(grid_all)

    with h5py.File(path, "r") as f:
        yield f


@pytest.fixture(scope="session")
def h5_out_1d(tmpdir_factory, grid_all):
    path = tmpdir_factory.mktemp("h5") / "gridadmin_1d.h5"

    grid_1d = copy(grid_all)
    grid_1d.quadtree_stats = None
    with GridAdminOut(path) as out:
        out.write(grid_1d)

    with h5py.File(path, "r") as f:
        yield f


# obtained from bergermeer gridadmin.h5, edited:
# - 13280 nodes to 3 nodes
# - int64 to int32
@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("bottom_level", (4,), "float64"),
        ("calculation_type", (4,), "int32"),
        ("cell_coords", (4, 4), "float64"),
        ("content_pk", (4,), "int32"),
        ("coordinates", (2, 4), "float64"),
        ("display_name", (4,), "|S64"),
        ("drain_level", (4,), "float64"),
        ("id", (4,), "int32"),
        ("initial_waterlevel", (4,), "float64"),
        ("is_manhole", (4,), "int32"),
        ("manhole_indicator", (4,), "int32"),
        ("node_type", (4,), "int32"),
        ("pixel_coords", (4, 4), "int32"),
        ("pixel_width", (4,), "int32"),
        ("shape", (4,), "|S4"),
        ("storage_area", (4,), "float64"),
        ("sumax", (4,), "float64"),
        ("surface_level", (4,), "float64"),
        ("width", (4,), "float64"),
        ("x_coordinate", (4,), "float64"),
        ("y_coordinate", (4,), "float64"),
        ("z_coordinate", (4,), "float64"),
        ("zoom_category", (4,), "int32"),
        ("code", (4,), "|S32"),  # added
        ("dmax", (4,), "float64"),  # added
        ("s1d", (4,), "float64"),  # added
        ("embedded_in", (4,), "int32"),  # added
        ("boundary_type", (4,), "int32"),  # added
        ("has_dem_averaged", (4,), "int32"),  # added
    ],
)
def test_write_nodes(h5_out, dataset, shape, dtype):
    assert h5_out["nodes"][dataset].shape == shape
    assert h5_out["nodes"][dataset].dtype == np.dtype(dtype)


@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("bottom_level", (3,), "float64"),
        ("calculation_type", (3,), "int32"),
        ("cell_coords", (4, 3), "float64"),
        ("content_pk", (3,), "int32"),
        ("coordinates", (2, 3), "float64"),
        ("display_name", (3,), "|S64"),
        ("drain_level", (3,), "float64"),
        ("id", (3,), "int32"),
        ("initial_waterlevel", (3,), "float64"),
        ("is_manhole", (3,), "int32"),
        ("manhole_indicator", (3,), "int32"),
        ("node_type", (3,), "int32"),
        ("pixel_coords", (4, 3), "int32"),
        ("pixel_width", (3,), "int32"),
        ("shape", (3,), "|S4"),
        ("storage_area", (3,), "float64"),
        ("sumax", (3,), "float64"),
        ("surface_level", (3,), "float64"),
        ("width", (3,), "float64"),
        ("x_coordinate", (3,), "float64"),
        ("y_coordinate", (3,), "float64"),
        ("z_coordinate", (3,), "float64"),
        ("zoom_category", (3,), "int32"),
        ("code", (3,), "|S32"),  # added
        ("dmax", (3,), "float64"),  # added
        ("s1d", (3,), "float64"),  # added
        ("embedded_in", (3,), "int32"),  # added
        ("boundary_id", (3,), "int32"),  # added
        ("boundary_type", (3,), "int32"),  # added
        ("has_dem_averaged", (3,), "int32"),  # added
    ],
)
def test_write_nodes_embedded(h5_out, dataset, shape, dtype):
    assert h5_out["nodes_embedded"][dataset].shape == shape
    assert h5_out["nodes_embedded"][dataset].dtype == np.dtype(dtype)


# obtained from bergermeer gridadmin.h5, edited:
# - 5 lines to 5 lines
# - int64 to int32
@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("calculation_type", (6,), "int32"),
        ("code", (6,), "|S32"),
        ("connection_node_end_pk", (6,), "int32"),
        ("connection_node_start_pk", (6,), "int32"),
        ("content_pk", (6,), "int32"),
        ("content_type", (6,), "|S10"),
        ("crest_level", (6,), "float64"),
        ("crest_type", (6,), "int32"),
        ("cross_section_height", (6,), "float64"),
        ("cross_section_shape", (6,), "int32"),
        ("cross_section_width", (6,), "float64"),
        ("discharge_coefficient", (6,), "float64"),
        ("discharge_coefficient_negative", (6,), "float64"),
        ("discharge_coefficient_positive", (6,), "float64"),
        ("display_name", (6,), "|S64"),
        ("dist_calc_points", (6,), "float64"),
        ("friction_type", (6,), "int32"),
        ("friction_value", (6,), "float64"),
        ("veg_coef1", (6,), "float64"),
        ("veg_coef2", (6,), "float64"),
        ("id", (6,), "int32"),
        ("invert_level_end_point", (6,), "float64"),
        ("invert_level_start_point", (6,), "float64"),
        ("kcu", (6,), "int32"),
        ("lik", (6,), "int32"),
        ("line", (2, 6), "int32"),
        ("cross_pix_coords", (4, 6), "int32"),
        ("line_coords", (4, 6), "float64"),
        ("line_geometries", (6,), "object"),
        ("material", (6,), "int32"),
        ("sewerage", (6,), "int32"),
        ("sewerage_type", (6,), "int32"),
        ("zoom_category", (6,), "int32"),
        ("s1d", (6,), "float64"),  # added
        ("ds1d", (6,), "float64"),  # added
        ("dpumax", (6,), "float64"),  # added
        ("flod", (6,), "float64"),  # added
        ("flou", (6,), "float64"),  # added
        ("cross1", (6,), "int32"),  # added
        ("cross2", (6,), "int32"),  # added
        ("cross_weight", (6,), "float64"),  # added
        ("windshieldings", (8, 6), "float64"),
    ],
)
def test_write_lines(h5_out, dataset, shape, dtype):
    assert h5_out["lines"][dataset].shape == shape
    assert h5_out["lines"][dataset].dtype == np.dtype(dtype)


def test_line_geometries(h5_out):
    # line geometries are stored as a variable-length array [x, x, ..., y, y, ...]
    data = h5_out["lines"]["line_geometries"][:]
    assert np.isnan(data[0]).all()
    assert data[1].tolist() == [1, 2, 1, 2]
    assert data[2].tolist() == [1, 2, 3, 1, 2, 3]


def test_line_cross_mapping(h5_out):
    # cross ids should be mapped to (1 based) cross indexes
    assert_equal(h5_out["lines"]["cross1"][1:], [2, -9999, 1, 1, 2])
    assert_equal(h5_out["lines"]["cross2"][1:], [-9999, -9999, -9999, -9999, 3])


@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("lgrmin", (), "int32"),
        ("kmax", (), "int32"),
        ("mmax", (2,), "int32"),
        ("nmax", (2,), "int32"),
        ("dx", (2,), "float64"),
        ("dxp", (), "float64"),
        ("x0p", (), "float64"),
        ("y0p", (), "float64"),
    ],
)
def test_write_quadtree(h5_out, dataset, shape, dtype):
    assert h5_out["grid_coordinate_attributes"][dataset].shape == shape
    assert h5_out["grid_coordinate_attributes"][dataset].dtype == np.dtype(dtype)


@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("lgrmin", (), "int32"),
        ("kmax", (), "int32"),
        ("mmax", (1,), "int32"),
        ("nmax", (1,), "int32"),
        ("dx", (1,), "float64"),
        ("dxp", (), "float64"),
        ("x0p", (), "float64"),
        ("y0p", (), "float64"),
    ],
)
def test_write_quadtree_pure_1d(h5_out_1d, dataset, shape, dtype):
    assert h5_out_1d["grid_coordinate_attributes"][dataset].shape == shape
    assert h5_out_1d["grid_coordinate_attributes"][dataset].dtype == np.dtype(dtype)


@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        # ('ijmax', (), 'i4'),  deprecated
        # ('imax', (), 'i4'),  deprecated
        ("infl1d", (), "i4"),
        ("ingrw1d", (), "i4"),
        # ('jap1d', (), 'i4'),  deprecated
        # ('jmax', (), 'i4'),  deprecated
        ("l1dtot", (), "i4"),
        ("l2dtot", (), "i4"),
        # ('levnms', (), 'i4'),  deprecated
        # ('lgrmin', (), 'i4'),  deprecated
        ("lgrtot", (), "i4"),
        ("lgutot", (), "i4"),
        ("lgvtot", (), "i4"),
        # ('linall', (), 'i4'),  deprecated (can be derived from others)
        # ('lintot', (), 'i4'),  deprecated (can be derived from others)
        ("liutot", (), "i4"),
        ("livtot", (), "i4"),
        ("n1dobc", (), "i4"),
        ("n1dtot", (), "i4"),
        # ('n2dall', (), 'i4'),  deprecated (can be derived from others)
        ("n2dobc", (), "i4"),
        ("n2dtot", (), "i4"),
        ("ngr2bc", (), "i4"),
        ("ngrtot", (), "i4"),
        ("nob2dg", (), "i4"),
        ("nob2ds", (), "i4"),
        # ('nodall', (), 'i4'),  deprecated (can be derived from others)
        # ('nodobc', (), 'i4'),  deprecated (can be derived from others)
        # ('nodtot', (), 'i4'),  deprecated (can be derived from others)
    ],
)
def test_write_meta(h5_out, dataset, shape, dtype):
    assert h5_out["meta"][dataset].shape == shape
    assert h5_out["meta"][dataset].dtype == np.dtype(dtype)


@pytest.mark.parametrize(
    "attr,shape,dtype",
    [
        ("epsg_code", (), "S"),
        ("crs_wkt", (), "S"),  # added
        ("has_1d", (), "bool"),  # changed to bool
        ("has_2d", (), "bool"),  # changed to bool
        ("has_0d", (), "bool"),
        ("extent_1d", (4,), "float64"),
        ("extent_2d", (4,), "float64"),
        ("zero_dim_extent", (4,), "float64"),
        ("has_breaches", (), "bool"),  # changed to bool
        ("has_groundwater", (), "bool"),  # changed to bool
        ("has_groundwater_flow", (), "bool"),  # changed to bool
        ("has_interception", (), "bool"),  # changed to bool
        ("has_pumpstations", (), "bool"),  # changed to bool
        ("has_simple_infiltration", (), "bool"),  # changed to bool
        ("has_interflow", (), "bool"),  # added
        ("has_initial_waterlevels", (), "bool"),  # added
        ("model_name", (), "S"),
        ("model_slug", (), "S"),
        ("revision_hash", (), "S"),
        ("revision_nr", (), "int32"),  # changed to int
        ("threedi_version", (), "S"),
        ("threedicore_version", (), "S"),
    ],
)
def test_write_attrs(h5_out, attr, shape, dtype):
    assert h5_out.attrs[attr].shape == shape
    actual = h5_out.attrs[attr].dtype
    if dtype == "S":
        actual = actual.char
    assert np.dtype(actual) == np.dtype(dtype)


# obtained from bergermeer gridadmin.h5, edited:
# - 20 pumps to 4
# - int64 to int32
@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("bottom_level", (4,), "float64"),
        ("capacity", (4,), "float64"),
        ("connection_node_end_pk", (4,), "int32"),
        ("connection_node_start_pk", (4,), "int32"),
        ("content_pk", (4,), "int32"),
        ("coordinates", (2, 4), "float64"),
        ("display_name", (4,), "|S64"),  # increased size from 24 to 64
        ("id", (4,), "int32"),
        ("lower_stop_level", (4,), "float64"),
        ("node1_id", (4,), "int32"),
        ("node2_id", (4,), "int32"),
        ("node_coordinates", (4, 4), "float64"),
        # ("nodp1d", (2, 3), "int32"), removed
        # ("p1dtyp", (4,), "int32"), removed
        ("start_level", (4,), "float64"),
        ("type", (4,), "int32"),
        ("zoom_category", (4,), "int32"),
        ("code", (4,), "|S32"),  # added
        ("upper_stop_level", (4,), "float64"),  # added
    ],
)
def test_write_pumps(h5_out, dataset, shape, dtype):
    assert h5_out["pumps"][dataset].shape == shape
    assert h5_out["pumps"][dataset].dtype == np.dtype(dtype)


@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("id", (4,), "int32"),
        ("code", (4,), "|S32"),  # added
        ("shape", (4,), "int32"),
        ("content_pk", (4,), "int32"),
        ("width_1d", (4,), "float64"),
        ("offset", (4,), "int32"),
        ("count", (4,), "int32"),
        ("tables", (2, 6), "float64"),
    ],
)
def test_write_cross_sections(h5_out, dataset, shape, dtype):
    assert h5_out["cross_sections"][dataset].shape == shape
    assert h5_out["cross_sections"][dataset].dtype == np.dtype(dtype)


# obtained from bergermeer gridadmin.h5
# note that there is no dummy element in this group (just 2 levees)
@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("id", (2,), "int32"),
        ("crest_level", (2,), "float64"),
        ("max_breach_depth", (2,), "float64"),
        ("coords", (2,), "object"),
    ],
)
def test_write_levees(h5_out, dataset, shape, dtype):
    assert h5_out["levees"][dataset].shape == shape
    assert h5_out["levees"][dataset].dtype == np.dtype(dtype)


def test_write_levees_coords(h5_out):
    # coords are stored as a variable-length array [x, x, ..., y, y, ...]
    data = h5_out["levees"]["coords"][:]
    assert data[0].tolist() == [1, 2, 4, 1, 2, 4]
    assert data[1].tolist() == [1, 2, 3, 1, 2, 3]


# obtained from bergermeer gridadmin.h5, edited:
# - 1 breach to 2
# - int64 to int32
@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("id", (3,), "int32"),
        ("levl", (3,), "int32"),
        ("levbr", (3,), "float64"),
        ("levmat", (3,), "int32"),
        ("content_pk", (3,), "int32"),
        ("coordinates", (2, 3), "float64"),
        ("code", (3,), "|S32"),  # added
        ("display_name", (3,), "|S64"),  # added
        # ("llev", (1931,), "int32"),  dropped
        # ("kcu", (2,), "int32"),  dropped
        # ("seq_ids", (2,), "int32"),  dropped
    ],
)
def test_write_breaches(h5_out, dataset, shape, dtype):
    assert h5_out["breaches"][dataset].shape == shape
    assert h5_out["breaches"][dataset].dtype == np.dtype(dtype)


@pytest.mark.parametrize(
    "group,dataset,expected",
    [
        ("nodes", "id", 1),
        ("nodes", "dmax", 1.2),
        ("nodes_embedded", "embedded_in", 2),
        ("nodes_embedded", "dmax", 2.3),
        ("lines", "id", 1),
        ("lines", "dpumax", 1.2),
        ("pumps", "id", 1),
        ("pumps", "capacity", 0.1),
        ("cross_sections", "id", 1),
        ("cross_sections", "width_1d", 0.2),
        ("cross_sections", "offset", 0),  # reference to tables dataset, not increased
        ("lines", "line", [1, 2]),  # reference to node
        ("lines", "cross1", 2),  # reference to cross section
        ("lines", "cross2", -9999),  # reference to cross section
        ("pumps", "node1_id", 1),  # reference to node
        ("pumps", "node2_id", 2),  # reference to node
        ("breaches", "id", 1),
        ("breaches", "levl", 4),  # reference to line
    ],
)
def test_not_off_by_one(h5_out, group, dataset, expected):
    # gridadmin contains a dummy element at index 0 (so index 1 is the first)
    # references should also be increased by one
    assert h5_out[group][dataset][..., 1].tolist() == expected


@pytest.mark.parametrize(
    "dataset,shape,dtype",
    [
        ("area", (3,), "float64"),
        ("cci", (4,), "int32"),
        ("centroid_x", (3,), "float64"),
        ("centroid_y", (3,), "float64"),
        ("code", (3,), "|S100"),
        ("display_name", (3,), "|S250"),
        ("dry_weather_flow", (3,), "float64"),
        ("fac", (4,), "float64"),
        ("fb", (3,), "float64"),
        ("fe", (3,), "float64"),
        ("function", (3,), "|S64"),
        ("id", (3,), "int32"),
        ("imp", (4,), "int32"),
        ("infiltration_flag", (3,), "bool"),
        ("ka", (3,), "float64"),
        ("kh", (3,), "float64"),
        ("nr_of_inhabitants", (3,), "float64"),
        ("nxc", (4,), "float64"),
        ("nyc", (4,), "float64"),
        ("outflow_delay", (3,), "float64"),
        ("pk", (4,), "int32"),
        ("storage_limit", (3,), "float64"),
    ],
)
def test_write_surface(h5_out, dataset, shape, dtype):
    assert h5_out["surface"][dataset].shape == shape
    assert h5_out["surface"][dataset].dtype == np.dtype(dtype)


def test_encoding(h5_out):
    assert h5_out["nodes"]["display_name"][1].decode() == "foo"
    assert h5_out["nodes"]["display_name"][2].decode() == "金蟾"
    assert h5_out["nodes"]["display_name"][3].decode() == ""



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from pyproj import CRS

from threedigrid_builder.exceptions import SchematisationError

__all__ = ["OutputInterface", "RasterInterface"]


class OutputInterface(ABC):
    """The metaclass (class template) for data output"""

    def __init__(self, path: Path):
        self.path = path
        super().__init__()

    @staticmethod
    def available():
        return True

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, type, value, traceback):
        pass

    @abstractmethod
    def write(self, grid):
        pass


class RasterInterface(ABC):
    """The metaclass (class template) for raster data input"""

    GT_TOLERANCE = 7

    def __init__(
        self,
        path: Path,
        model_area_path: Optional[Path] = None,
    ):
        self.path = path
        self.model_area_path = model_area_path
        self.transform = None
        self.crs = None
        super().__init__()

    def set_transform(self, transform):
        """Transform should be given in RasterIO order:

        >>> dx, 0, x, 0, dy, y
        """
        a, b, c, d, e, f = [round(x, self.GT_TOLERANCE) for x in transform]

        # check if pixels are square and if cross terms are 0
        if abs(a) != abs(e):
            raise SchematisationError(
                f"Raster pixels are non-square ({abs(a)}x{abs(e)})."
            )
        if b != 0.0 or d != 0.0:
            raise SchematisationError(
                f"Raster pixel grid has tilt (cross terms: {b},{d})."
            )

        self.transform = (a, b, c, d, e, f)

    def set_crs(self, wkt: str):
        self.crs = CRS.from_wkt(wkt)

    @property
    def pixel_size(self):
        return abs(self.transform[0])

    def __repr__(self):
        return f"<{self.__class__.__name__} of file {self.path}>"

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, type, value, traceback):
        pass

    @abstractmethod
    def read(self):
        pass



import numpy as np
import shapely

from threedigrid_builder.base import Array, PointsOnLine
from threedigrid_builder.constants import FrictionType

__all__ = ["CrossSectionLocations"]


class CrossSectionLocation:
    id: int
    code: str
    the_geom: shapely.Geometry
    definition_id: id  # refers to CrossSectionDefinition
    channel_id: id  # refers to Channel
    reference_level: float
    bank_level: float
    friction_type: FrictionType
    friction_value: float
    vegetation_stem_density: float
    vegetation_stem_diameter: float
    vegetation_height: float
    vegetation_drag_coefficient: float


class CrossSectionLocations(Array[CrossSectionLocation]):
    def apply_to_lines(self, lines, channels, extrapolate=False):
        """Apply cross section locations to lines

        Args:
            lines (Lines): Changed inplace
            channels (Channels)
            extrapolate (bool): Whether to allow extrapolation. Extrapolation may occur
                when there are multiple cross section locations and some line centers
                are not in between them. When turned off, the values of those lines
                become equal to the values of the closest cross section location.
                Default False.

        The following lines attributes are changed inplace:

        - cross_id1: the id of the first cross section definition
        - cross_id2: the id of the second cross section definition
        - cross_weight: the weight of the first cross section definition
        - frict_type1: the friction type of the first cross section location
        - frict_type2: the friction type of the second cross section location
        - frict_value1: the friction value of the first cross section location
        - frict_value2: the friction value of the second cross section location
        - veg_coef1: the product of vegetation properties of the first cross section location (except height)
        - veg_coef2: the product of vegetation properties of the second cross section location (except height)
        - veg_height1: the vegetation height of the first cross section
        - veg_height2: the vegetation height of the second cross section
        - invert_level_start_point: 'reference_level' interpolated at the line end
        - invert_level_end_point: 'reference_level' interpolated at the line start
        - dpumax: the largest of the two invert levels
        """
        # Mask the lines to only the Channel lines
        cross_loc1, cross_loc2, cross_weight = compute_weights(
            lines.content_pk,
            lines.s1d,
            self,
            channels,
            extrapolate=extrapolate,
        )

        idx1 = self.id_to_index(cross_loc1)
        idx2 = self.id_to_index(cross_loc2)
        lines.cross_id1 = self.definition_id[idx1]
        lines.cross_id2 = self.definition_id[idx2]
        lines.cross_weight = cross_weight

        lines.frict_type1 = self.friction_type[idx1]
        lines.frict_type2 = self.friction_type[idx2]
        lines.frict_value1 = self.friction_value[idx1]
        lines.frict_value2 = self.friction_value[idx2]

        lines.veg_coef1 = (
            self.vegetation_stem_density[idx1]
            * self.vegetation_stem_diameter[idx1]
            * self.vegetation_drag_coefficient[idx1]
        )
        lines.veg_coef1[np.isnan(lines.veg_coef1)] = 0.0

        lines.veg_coef2 = (
            self.vegetation_stem_density[idx2]
            * self.vegetation_stem_diameter[idx2]
            * self.vegetation_drag_coefficient[idx2]
        )
        lines.veg_coef2[np.isnan(lines.veg_coef2)] = 0.0
        lines.veg_height1 = self.vegetation_height[idx1]
        lines.veg_height1[np.isnan(lines.veg_height1)] = 0.0
        lines.veg_height2 = self.vegetation_height[idx2]
        lines.veg_height2[np.isnan(lines.veg_height2)] = 0.0

        # Compute invert levels and start and end
        lines.invert_level_start_point = compute_bottom_level(
            lines.content_pk,
            lines.s1d - (lines.ds1d / 2),
            self,
            channels,
            extrapolate=extrapolate,
        )
        lines.invert_level_end_point = compute_bottom_level(
            lines.content_pk,
            lines.s1d + (lines.ds1d / 2),
            self,
            channels,
            extrapolate=extrapolate,
        )
        lines.dpumax = np.maximum(
            lines.invert_level_start_point, lines.invert_level_end_point
        )


def compute_weights(channel_id, points, cs, channels, extrapolate):
    """Compute cross section weights for points on channels.

    Points on channels are specified with their channel id and the distance along that
    channel id.

    Args:
        channel_id (ndarray of int): The channel ids for which to compute weights. Each
            id may occur multiple times, if there are multiple points on that channel
            to compute the weights on. This array must be in ascending order.
        points (ndarray of float): The location of the points to compute the weights for,
            measured as a distance along the channel. Array must be the same size as
            channel_id and ordered (per channel) in ascending order.
        cs (CrossSectionLocations)
        channels (Channels): Used to lookup the channel geometry
        extrapolate (bool): Whether to allow the weight to go outside of [0, 1]

    Returns:
        tuple of cross_loc1, cross_loc2, cross_weight

    See also:
        Channels.get_lines: computes the lines in the correct order
    """
    if not np.in1d(channels.id, cs.channel_id).all():
        missing = set(channels.id) - set(cs.channel_id)
        raise ValueError(f"Channels {missing} have no cross section location set.")

    if len(channel_id) == 0:
        return (
            np.empty((0,), dtype=int),
            np.empty((0,), dtype=int),
            np.empty((0,), dtype=float),
        )

    cs_points = PointsOnLine.from_geometries(
        channels.linestrings,
        cs.the_geom,
        channels.id_to_index(cs.channel_id),
        content_pk=cs.id,
    )
    queried_points = PointsOnLine.from_s1d(
        channels.linestrings, points, channels.id_to_index(channel_id)
    )

    # Find what CS location comes after and before each midpoint
    cs_idx_1, cs_idx_2 = cs_points.neighbours(queried_points)

    # Map index to id and create the array that matches the input channel_id and ds
    cross_loc1 = cs_points.content_pk[cs_idx_1]
    cross_loc2 = cs_points.content_pk[cs_idx_2]

    # Compute the weights. For each line, we have 3 times s:
    # 1. the s of the CrossSectionLocation before it (cs_idx_1)
    # 2. the s of the CrossSectionLocation after it (cs_idx_2)
    # 3. the s of the midpoint (velocity point) (queried_points)
    #
    # The weight is calculated such that a value at midpoint can be interpolated
    # as:   weight * v_1 + (1 - weight) * v_2
    s_1 = cs_points.s1d[cs_idx_1]
    s_2 = cs_points.s1d[cs_idx_2]
    with np.errstate(divide="ignore", invalid="ignore"):
        # this transforms 1 / 0 to inf and 0 / 0 to nan without warning
        cross_weight = (s_2 - queried_points.s1d) / (s_2 - s_1)
    cross_weight[~np.isfinite(cross_weight)] = 1.0

    if not extrapolate:
        cross_weight = np.clip(cross_weight, 0.0, 1.0)

    return cross_loc1, cross_loc2, cross_weight


def interpolate(cross_loc1, cross_loc2, cross_weight, cs, cs_attr="reference_level"):
    """Interpolate an attribute of CrossSectionLocation using known weights.

    Args:
        cross_loc1 (ndarray of int): see compute_weights
        cross_loc2 (ndarray of int): see compute_weights
        cross_weight (ndarray of float): see compute_weights
        cs (CrossSectionLocations)
        cs_attr ({"reference_level", "bank_level"})

    Returns:
        an array of the same shape cross_loc1 containing the interpolated values

    See also:
        compute_weights: computes the interpolation weights
    """
    values = getattr(cs, cs_attr)
    left = values.take(cs.id_to_index(cross_loc1))
    right = values.take(cs.id_to_index(cross_loc2))
    return cross_weight * left + (1 - cross_weight) * right


def compute_bottom_level(
    channel_id, ds, cs, channels, cs_attr="reference_level", extrapolate=False
):
    """Compute levels by interpolating/extrapolating between cross sections

    This can be used at nodes (for dmax) or at line centres (for dpumax).

    Args:
        channel_id (ndarray of int): see compute_weights
        ds (ndarray of float): see compute_weights
        cs (CrossSectionLocations): the reference_level is inter/extrapolated
        channels (Channels): see compute_weights
        cs_attr ({"reference_level", "bank_level"})
        extrapolate (bool): Whether to allow extrapolation. Extrapolation may occur
            when there are multiple cross section locations and some points (ds)
            are not in between them. When turned off, the values at those points
            become equal to the values of the closest cross section locations.
            Default False.

    Returns:
        an array of the same shape as channel_id containing the interpolated values

    See also:
        compute_weights: computes the interpolation weights
    """
    cross_loc1, cross_loc2, cross_weight = compute_weights(
        channel_id, ds, cs, channels, extrapolate=extrapolate
    )
    return interpolate(cross_loc1, cross_loc2, cross_weight, cs, cs_attr)



from typing import Tuple

from .array import Array

__all__ = ["Quarters"]


class Quarter:
    id: int
    line: Tuple[int, int]
    neighbour_node: Tuple[int, int]


class Quarters(Array[Quarter]):
    """Calculation quarters."""

    pass



import numpy as np

from threedigrid_builder.constants import ContentType

from .array import Array
from .lines import Lines
from .nodes import Nodes

__all__ = ["LineHalfs"]


class LineHalf:
    id: int
    line_idx: int
    is_start: bool
    node_id: int


class LineHalfs(Array[LineHalf]):
    """Each line has 2 line_halfs.

    This class makes it easier to reason about all the objects on a node.
    """

    scalars = ("lines",)

    @classmethod
    def from_nodes_lines(
        cls,
        nodes: Nodes,
        lines: Lines,
        line_mask=None,
        node_mask=None,
    ) -> "LineHalfs":
        """Return the line_halfs, optionally filtering with node / line masks."""
        if line_mask is None:
            where = np.arange(len(lines), dtype=int)
        else:
            where = np.where(line_mask)[0]
        n = len(where)
        result = cls(
            lines=lines,
            id=range(n * 2),
            line_idx=np.repeat(where, 2),
            node_id=lines.line[where].ravel(),
            is_start=np.tile([True, False], n),
        )
        if node_mask is not None:
            result = result[np.isin(result.node_id, nodes.id[node_mask])]
        return result

    @classmethod
    def for_connection_nodes(
        cls,
        nodes: Nodes,
        lines: Lines,
        line_mask=None,
        node_mask=None,
    ) -> "LineHalfs":
        """Return the line_halfs for connection nodes

        Optionally filter by the content type of the lines.
        """
        is_cn = nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES
        node_mask = is_cn if node_mask is None else is_cn & node_mask
        return cls.from_nodes_lines(nodes, lines, line_mask, node_mask)

    @property
    def is_end(self):
        return ~self.is_start

    @property
    def invert_level(self):
        return np.where(
            self.is_start, self.invert_level_start_point, self.invert_level_end_point
        )

    @property
    def ds1d(self):
        return np.where(
            self.is_start,
            self.lines.ds1d_half[self.line_idx],
            self.lines.ds1d[self.line_idx] - self.lines.ds1d_half[self.line_idx],
        )

    def __getattr__(self, name):
        try:
            return super(self.__class__, self).__getattr__(name)
        except AttributeError:
            return getattr(self.lines, name)[self.line_idx]

    def _get_reduce_indices(self):
        """Return indices into self containing the same node ids.

        Intended for usage in reduction functions (see reduce).

        For instance:

         - self.node_id = [1, 1, 5, 5, 5]
         - gives: indices = [0, 2]
        """
        if len(self) == 0:
            return np.empty((0,), dtype=int)
        diff = np.diff(self.node_id)
        if np.any(diff < 0):
            raise ValueError("line_halfs must be ordered by node_id")
        return np.concatenate([[0], np.where(diff > 0)[0] + 1])

    def get_reduce_id(self):
        """Return unique node ids.

        Intended for usage with reduction functions (see reduce).
        """
        return self.node_id[self._get_reduce_indices()]

    def reduce(self, reduceat, values):
        """Compute the per-node reduction of 'values'

        The 'reduceat' function is normally a numpy ufunc.reduceat. This function
        is expected to reduce 'values' on index ranges contained in 'indices'. The length
        of 'indices' must equal the length of the returned value.

        For instance:

         - self.node_id = [1, 1, 5, 5, 5]
         - values = [1., 3., 5., 7., 9.]
         - gives: indices = [0, 2]
         - and reduceat is expected to act on index ranges [0, 2) and [2, <end>)

        Returns node ids and corresponding maximum values.
        """
        values = np.asarray(values)
        if values.shape[0] != len(self):
            raise ValueError("values must have the same length as self")
        if len(values) == 0:
            return np.empty((0,) + values.shape[1:])
        indices = self._get_reduce_indices()
        return reduceat(values, indices)

    def nanmin(self, values):
        return self.reduce(np.fmin.reduceat, values)

    def sum(self, values):
        return self.reduce(np.add.reduceat, values)

    def first(self, values):
        return self.reduce(lambda x, y: x[y], values)



import itertools

import numpy as np
import pytest
import shapely
from numpy.testing import assert_array_equal

from threedigrid_builder.base import Nodes
from threedigrid_builder.grid import Channels, ConnectionNodes, CrossSectionLocations


@pytest.fixture
def channels():
    return Channels(
        the_geom=shapely.linestrings([[(0, 0), (6, 0), (6, 6)]]),
        dist_calc_points=np.array([5.0]),
        id=np.array([1]),
        code=np.array(["one"]),
        connection_node_start_id=np.array([21]),
        connection_node_end_id=np.array([42]),
        calculation_type=np.array([2]),
    )


def test_get_1d2d_exchange_levels(channels):
    locations = CrossSectionLocations(
        id=[2, 5],
        the_geom=shapely.points([(0, 0), [6, 6]]),
        bank_level=[1.0, 13.0],
        channel_id=[1, 1],
    )

    actual = channels.get_1d2d_exchange_levels(
        content_pk=np.array([1, 1]), s1d=np.array([3.0, 6.0]), locations=locations
    )

    # bank levels are interpolated
    assert_array_equal(actual, [4.0, 7.0])


def test_is_closed(channels):
    actual = channels.is_closed(np.array([1, 3]))

    assert_array_equal(actual, [False, False])


@pytest.mark.parametrize(
    "thickness,hc_out,hc_in,expected",
    [
        (0.1, 3.0, 2.0, True),
        (0.0, 3.0, 2.0, False),
        (np.nan, 3.0, 2.0, False),
        (0.1, np.nan, 2.0, False),
        (0.1, 3.0, np.nan, False),
    ],
)
def test_has_groundwater_exchange(thickness, hc_out, hc_in, expected):
    channels = Channels(
        id=[1],
        exchange_thickness=thickness,
        hydraulic_conductivity_out=hc_out,
        hydraulic_conductivity_in=hc_in,
    )

    actual = channels.has_groundwater_exchange

    assert len(actual) == 1
    assert actual[0] == expected


@pytest.mark.parametrize(
    "thickness,hc_out,hc_in,res_out,res_in",
    [
        (0.1, 3.0, 2.0, 30.0, 20.0),
        (np.nan, 3.0, 2.0, np.nan, np.nan),
        (0.1, np.nan, 2.0, np.nan, 20.0),
        (0.1, 3.0, np.nan, 30.0, np.nan),
    ],
)
def test_set_hydraulic_resistance(channels, thickness, hc_out, hc_in, res_in, res_out):
    channels.exchange_thickness = np.array([thickness])
    channels.hydraulic_conductivity_out = np.array([hc_out])
    channels.hydraulic_conductivity_in = np.array([hc_in])

    cn = ConnectionNodes(id=[21, 42], the_geom=shapely.points([(0, 21), (1, 25)]))
    nodes = Nodes(id=[10], content_pk=[1], s1d=[0.5])
    lines = channels.get_lines(cn, nodes, itertools.count(start=0))

    assert_array_equal(lines.hydraulic_resistance_in, [res_in, res_in])
    assert_array_equal(lines.hydraulic_resistance_out, [res_out, res_out])



import numpy as np
import shapely

from .array import Array

__all__ = ["LineStrings", "PointsOnLine", "LinesOnLine"]


COORD_EQUAL_ATOL = 1e-8  # the distance below which coordinates are considered equal


class PointOnLine:
    """A point determined by its position 'along' a linestring"""

    id: int  # just a unique number, mostly unused
    content_pk: int  # externally determined id
    s1d: float  # the position along the linestring
    linestring_idx: int
    secondary_content_pk: int  # for double connected points


class LineOnLine:
    """A line determined by its position 'along' a linestring"""

    id: int  # just a unique number, mostly unused
    content_pk: int  # externally determined id
    s1d_start: float  # the position along the linestring
    s1d_end: float  # the position along the linestring
    linestring_idx: int


class LineStrings:
    def __init__(self, objects):
        geom_types = shapely.get_type_id(objects.the_geom)
        assert np.all((geom_types == -1) | (geom_types == 1))
        self.objects = objects

    def __len__(self):
        return len(self.objects)

    def __getattr__(self, name):
        return getattr(self.objects, name)

    @property
    def length(self):
        return shapely.length(self.the_geom)

    @property
    def length_cumulative(self):
        result = np.zeros(len(self) + 1, dtype=float)
        np.cumsum(self.length + 1e-6, out=result[1:])
        return result

    def segmentize(self, points: "PointsOnLine") -> "LinesOnLine":
        """Return lines that result from splitting self at 'points'"""
        segment_counts = np.bincount(points.linestring_idx, minlength=len(self)) + 1

        # cut the channel geometries into segment geometries
        start_s, end_s, segment_idx = segment_start_end(
            self.the_geom, segment_counts, points.s1d
        )
        return LinesOnLine(
            linestrings=self,
            id=range(len(segment_idx)),
            s1d_start=start_s,
            s1d_end=end_s,
            linestring_idx=segment_idx,
        )

    def sanitize(self, points_1, points_2) -> None:
        """Make sure that the self go from points_1 to points_2.

        The linestring is reversed if the distance from the linestring start to point_1
        and v.v. is lowered.

        Then, point_1 and/or point_2 are added to the linestring if the distance is larger
        than the tolerance of 1e-8.

        This function acts inplace on self.
        """
        has_geom = np.where(shapely.is_geometry(self.the_geom))[0]
        if len(has_geom) == 0:
            return

        # reverse the geometries where necessary
        node_start = points_1[has_geom]
        node_end = points_2[has_geom]
        line_start = shapely.get_point(self.the_geom[has_geom], 0)
        line_end = shapely.get_point(self.the_geom[has_geom], -1)

        dist_start_start = shapely.distance(node_start, line_start)
        dist_end_end = shapely.distance(node_end, line_end)
        dist_start_end = shapely.distance(node_start, line_end)
        dist_end_start = shapely.distance(node_end, line_start)
        needs_reversion = has_geom[
            (dist_start_start > dist_start_end) & (dist_end_end > dist_end_start)
        ]
        if len(needs_reversion) > 0:
            self.the_geom[needs_reversion] = shapely.reverse(
                self.the_geom[needs_reversion]
            )
            line_start[needs_reversion] = shapely.get_point(
                self.the_geom[needs_reversion], 0
            )
            line_end[needs_reversion] = shapely.get_point(
                self.the_geom[needs_reversion], -1
            )
            dist_start_start[needs_reversion] = dist_start_end[needs_reversion]
            dist_end_end[needs_reversion] = dist_end_start[needs_reversion]

        # for the remaining geometries; add point if necessary
        add_first = has_geom[dist_start_start > COORD_EQUAL_ATOL]
        if len(add_first) > 0:
            self.the_geom[add_first] = prepend_point(
                self.the_geom[add_first], points_1[add_first]
            )
        add_end = has_geom[dist_end_end > COORD_EQUAL_ATOL]
        if len(add_end) > 0:
            self.the_geom[add_end] = append_point(
                self.the_geom[add_end], points_2[add_end]
            )


class PointsOnLine(Array[PointOnLine]):
    scalars = ("linestrings",)

    @property
    def linestring_id(self):
        return self.linestrings.index_to_id(self.linestring_idx)

    @classmethod
    def empty(cls, linestrings: LineStrings):
        return cls(linestrings=linestrings, id=[])

    @classmethod
    def from_geometries(
        cls, linestrings: LineStrings, points, linestring_idx, **kwargs
    ):
        s1d = shapely.line_locate_point(linestrings.the_geom[linestring_idx], points)
        return cls.from_s1d(linestrings, s1d, linestring_idx, **kwargs)

    @classmethod
    def from_s1d(cls, linestrings: LineStrings, s1d, linestring_idx, **kwargs):
        if np.any(~np.isfinite(s1d)):
            raise ValueError("NaN values encountered in s1d")
        result = cls(
            linestrings=linestrings,
            id=np.arange(len(linestring_idx)),
            s1d=s1d,
            linestring_idx=linestring_idx,
            **kwargs,
        )
        result.reorder(np.argsort(result.s1d_cum))
        return result

    @property
    def s1d_cum(self):
        return self.s1d + self.linestrings.length_cumulative[self.linestring_idx]

    @property
    def the_geom(self):
        return shapely.line_interpolate_point(
            self.linestrings.the_geom[self.linestring_idx], self.s1d
        )

    @property
    def at_start(self):
        return self.s1d == 0.0

    @property
    def at_end(self):
        return self.s1d == self.linestrings.length[self.linestring_idx]

    def merge_with(self, other: "PointsOnLine"):
        if len(other) == 0:
            return self
        if len(self) == 0:
            first_i = 1
        else:
            first_i = self.id[-1] + 1
        other = other[:]
        other.id[:] = np.arange(first_i, first_i + len(other))
        result = self + other
        result.reorder(np.argsort(result.s1d_cum))
        return result

    def neighbours(self, other: "PointsOnLine"):
        """Compute neighbours in self of 'other'

        Returns two arrays that index into self: the neighbours 'before' and the
        neighbours 'after'.

        'other' must contain at least 1 point per linestring.

        If a point does not have a neigbour before it, the two returned indices
        will be the two points after it. Vice versa, if it does not have a neighbour after
        it, the two returned indices will be that of the neighboar before.

        If there there is only 1 point on a linestring, the neighbours will be equal.
        """
        idx_2 = np.searchsorted(self.s1d_cum, other.s1d_cum)
        idx_1 = idx_2 - 1
        out_of_bounds_1 = idx_1 < 0
        out_of_bounds_2 = idx_2 >= len(self)
        idx_1[out_of_bounds_1] = 0
        idx_2[out_of_bounds_2] = len(self) - 1

        # Fix situations where idx_1 is incorrect
        extrap_mask = (
            self.linestring_idx[idx_1] != other.linestring_idx
        ) | out_of_bounds_1
        idx_1[extrap_mask] = idx_2[extrap_mask]
        idx_2[extrap_mask] = np.clip(idx_2[extrap_mask] + 1, None, len(self) - 1)
        equalize = extrap_mask & (self.linestring_idx[idx_2] != other.linestring_idx)
        idx_2[equalize] -= 1

        # Fix situations where idx_2 is incorrect
        extrap_mask = (
            self.linestring_idx[idx_2] != other.linestring_idx
        ) | out_of_bounds_2
        idx_2[extrap_mask] = idx_1[extrap_mask]
        idx_1[extrap_mask] = np.clip(idx_2[extrap_mask] - 1, 0, None)
        equalize = extrap_mask & (self.linestring_idx[idx_1] != other.linestring_idx)
        idx_1[equalize] += 1

        return idx_1, idx_2


class LinesOnLine(Array[LineOnLine]):
    scalars = ("linestrings",)

    @property
    def s1d(self):
        return (self.s1d_start + self.s1d_end) / 2

    @property
    def ds1d(self):
        return self.s1d_end - self.s1d_start

    @property
    def the_geom(self):
        return line_substring(
            self.linestrings.the_geom, self.s1d_start, self.s1d_end, self.linestring_idx
        )

    def interpolate_points(self, desired_segment_size):
        """Compute points that divide sublinestrings into segments of equal length.

        Args:
            desired_segment_size (ndarray of float): the desired size of the segments; the
            actual size will depend on the linestring length and is computed by rounding
            ``line length / size`` to the nearest integer.
            Inf inputs will lead to no segmentation for that line (n_segments=1)

        Returns:
            PointsOnLine (excluding linestrings start and end)
        """
        # compute number of nodes to add per channel
        length = self.ds1d
        n_segments = np.maximum(np.round(length / desired_segment_size).astype(int), 1)
        segment_size = length / n_segments
        n_nodes = n_segments - 1

        # get the distance to the start of each channel
        i = counts_to_row_index(n_nodes)  # e.g. [0, 0, 0, 1, 1, 3]
        j = counts_to_column_index(n_nodes)  # e.g. [0, 1, 2, 0, 1, 0]
        dist_to_start = (j + 1) * segment_size[i] + self.s1d_start[i]

        return PointsOnLine.from_s1d(
            self.linestrings, dist_to_start, self.linestring_idx[i]
        )


def counts_to_ranges(counts):
    """Convert an array of list-of-lists counts to ranges.

    ``counts`` define lengths of lists that are concatenated into a 1D array.
    The output ranges index into that array. For example, ``arr[start[i]:stop[i]]`` will
    give all values belonging to list `i`.

    Args:
        counts (ndarray of int): list lengths

    Returns:
        tuple of start (ndarray of int), stop (ndarray of int)

    Example:
    >>> counts_to_ranges([3, 2, 0, 1, 0])
    (array([0, 3, 5, 5, 6]), array([3, 5, 5, 6, 6]))
    """
    if len(counts) == 0:
        return np.empty((0,), dtype=int), np.empty((0,), dtype=int)
    stop = np.cumsum(counts)
    start = np.roll(stop, 1)
    start[0] = 0
    return start, stop


def counts_to_row_index(counts):
    """Convert an array of list-of-lists counts into row indices into a 2D array.

    ``counts`` define lengths of lists that are concatenated into a 1D array. The output
    of this function assigns a row index to every element in the 1D array. The row index
    is counting over the outer list.

    Args:
        counts (ndarray of int): list lengths

    Returns:
        row index (ndarray of int)

    Example:
    >>> counts_to_row_index([3, 2, 0, 1, 0])
    array([0, 0, 0, 1, 1, 3])
    """
    (n,) = counts.shape
    return np.repeat(np.arange(n), counts)


def counts_to_column_index(counts):
    """Convert an array of list-of-lists counts into row indices into a 2D array.

    ``counts`` define lengths of lists that are concatenated into a 1D array. The output
    of this function assigns a column index to every element in the 1D array. The column
    index is counting over elements in a single list.

    Args:
        counts (ndarray of int): list lengths

    Returns:
        column index (ndarray of int)

    Example:
    >>> counts_to_indices([3, 2, 0, 1, 0])
    array([0, 1, 2, 0, 1, 0])
    """
    if len(counts) == 0:
        return np.empty((0,), dtype=int)
    start, stop = counts_to_ranges(counts)
    return np.arange(stop[-1]) - np.repeat(start, counts)


def line_substring(linestrings, start, end, index=None):
    """Divide linestrings into segments given [start, end] measured along the line.

    Note that there is no bound check done on start and end. Expect bogus results or
    errors if start is negative or if end is larger than the linestring length.

    Args:
        linestrings (ndarray of shapely.Geometry): Linestrings to segmentize
        start (ndarray of float): The start of the segment, measured along the line.
        end (ndarray of float): The end of the segment, measured along the line.
        index (ndarray of int, optional): An optional linestring index per start/end.
           If not given, then all input arrays should be equal sized.

    Returns:
        segments: the segments (sublinestrings of the input linestrings)
    """
    (n_lines,) = linestrings.shape
    if index is None:
        n_segments = n_lines
        index = np.arange(n_lines)
    else:
        (n_segments,) = index.shape

    if n_segments == 0:
        return np.empty((0,), dtype=object)

    coords, coord_line_idx = shapely.get_coordinates(linestrings, return_index=True)
    line_n_coords = shapely.get_num_coordinates(linestrings)

    ## compute the length of each vertex-vertex distance
    coord_ds = np.sqrt(np.sum((coords - np.roll(coords, 1, axis=0)) ** 2, axis=1))
    coord_s = np.cumsum(coord_ds)
    # compute corresponding length of each given start and end
    line_first_coord_idx, line_end_coord_idx = counts_to_ranges(line_n_coords)
    start_s = start + coord_s[line_first_coord_idx][index]
    end_s = end + coord_s[line_first_coord_idx][index]

    ## interpolate the start & end points along the linestrings
    start_insert_at = np.searchsorted(coord_s, start_s - COORD_EQUAL_ATOL)
    end_insert_at = np.searchsorted(coord_s, end_s - COORD_EQUAL_ATOL)
    # find points that would be located exactly at an existing coordinate
    start_eq_coord = np.abs(coord_s[start_insert_at] - start_s) < COORD_EQUAL_ATOL
    end_eq_coord = np.abs(coord_s[end_insert_at] - end_s) < COORD_EQUAL_ATOL
    # insert the existing coords and interpolate the others
    start_coords = np.empty((n_segments, 2))
    end_coords = np.empty((n_segments, 2))
    start_coords[start_eq_coord] = coords[start_insert_at[start_eq_coord]]
    end_coords[end_eq_coord] = coords[end_insert_at[end_eq_coord]]
    for i in range(2):
        start_coords[~start_eq_coord, i] = np.interp(
            start_s[~start_eq_coord], coord_s, coords[:, i]
        )
        end_coords[~end_eq_coord, i] = np.interp(
            end_s[~end_eq_coord], coord_s, coords[:, i]
        )

    ## fill an array of coordinates for the segmetns
    # the number of coordinates per segment are:
    # + 2 (line start and end)
    # + coordinates in between (end_insert_at - start_insert_at)
    # + correction where start_eq_coord (then there is 1 less "in between")
    # + no correction where end_eq_coord
    segment_n_coords = 2 + end_insert_at - start_insert_at - start_eq_coord
    segment_first_coord_idx, segment_end_coord_idx = counts_to_ranges(segment_n_coords)

    # fill in the start & end coords
    segment_coords = np.empty((segment_n_coords.sum(), 2))
    segment_coords[segment_first_coord_idx] = start_coords
    segment_coords[segment_end_coord_idx - 1] = end_coords

    # fill the remaining segment coordinates with original coordinates
    i = counts_to_row_index(segment_n_coords - 2)  # e.g. [0, 0, 0, 1, 1, 3]
    j = counts_to_column_index(segment_n_coords - 2)  # e.g. [0, 1, 2, 0, 1, 0]
    segment_coords_to_fill = segment_first_coord_idx[i] + j + 1
    coords_to_fill_with = start_insert_at[i] + start_eq_coord[i] + j
    segment_coords[segment_coords_to_fill] = coords[coords_to_fill_with]

    # construct the segments
    segments = shapely.linestrings(
        segment_coords,
        indices=counts_to_row_index(segment_n_coords),
    )

    return segments


def segment_start_end(linestrings, segment_counts, dist_to_start):
    """Utility function to use the output of segmentize as input of line_substrings.

    Args:
        linestrings (ndarray of shapely.Geometry): linestrings to segmentize
        segment_counts (ndarray of int): the number of segments per linestring
        dist_to_start (ndarray of float): the location of added nodes, measured along
          the linestring. The length of this array is such that:
          ``len(linestrings) + len(dist_to_start) = segment_counts.sum()``

    Returns:
        tuple of:
        - segment_start (ndarray of float): The segment start, measured along the line.
        - segment_end (ndarray of float): The segment end, measured along the line.
        - segment_index (ndarray of int): Indices mapping segments to input linestrings.
    """
    start_idx, last_idx = counts_to_ranges(segment_counts)
    last_idx -= 1

    start_s = np.full(segment_counts.sum(), np.nan)
    end_s = np.full(segment_counts.sum(), np.nan)
    # every first start_s is 0.0, the rest are added nodes with an s1d
    start_s[start_idx] = 0.0
    start_s[np.isnan(start_s)] = dist_to_start
    # every last end_s equals to the length, the rest are added nodes with an s1d
    end_s[last_idx] = shapely.length(linestrings)
    end_s[np.isnan(end_s)] = dist_to_start
    return start_s, end_s, counts_to_row_index(segment_counts)


def _add_point(linestrings, points, at=0):
    """Append or prepend points to linestrings"""
    counts = shapely.get_num_coordinates(linestrings)
    coords, index = shapely.get_coordinates(linestrings, return_index=True)
    start, end = counts_to_ranges(counts)
    if at == 0:
        insert_at = start
    elif at == -1:
        insert_at = end
    else:
        raise ValueError(f"Cannot insert at {at}")
    new_coords = np.insert(coords, insert_at, shapely.get_coordinates(points), axis=0)
    new_index = np.insert(index, insert_at, np.arange(len(insert_at)))
    return shapely.linestrings(new_coords, indices=new_index)


def prepend_point(linestrings, points):
    return _add_point(linestrings, points, at=0)


def append_point(linestrings, points):
    return _add_point(linestrings, points, at=-1)



import itertools
import logging
from typing import Iterator

import numpy as np
import shapely

from threedigrid_builder.base import LineHalfs, Lines, Nodes, replace, search
from threedigrid_builder.constants import CalculationType, ContentType, LineType
from threedigrid_builder.exceptions import SchematisationError

from .connection_nodes import ConnectionNodes
from .exchange_lines import ExchangeLines
from .obstacles import Obstacles
from .potential_breaches import PotentialBreaches

logger = logging.getLogger(__name__)

__all__ = ["Lines1D2D"]


class Lines1D2D(Lines):
    @classmethod
    def create(cls, nodes: Nodes, line_id_counter: Iterator[int]) -> "Lines1D2D":
        """Create the 1D-2D lines

        Sets: id, line[:, 1], content_type, content_pk
        """
        is_single = nodes.calculation_type == CalculationType.CONNECTED
        is_double = nodes.calculation_type == CalculationType.DOUBLE_CONNECTED
        node_idx = np.concatenate(
            [np.where(is_single | is_double)[0], np.where(is_double)[0]]
        )
        node_idx.sort()

        # Merge the arrays
        line = np.full((len(node_idx), 2), -9999, dtype=np.int32)
        line[:, 1] = nodes.index_to_id(node_idx)
        return cls(
            id=itertools.islice(line_id_counter, len(node_idx)),
            line=line,
            content_type=nodes.content_type[node_idx],
            content_pk=nodes.content_pk[node_idx],
        )

    @property
    def side_2d(self):
        return shapely.points(self.line_coords[:, :2])

    @property
    def side_1d(self):
        return shapely.points(self.line_coords[:, 2:])

    def assign_connection_nodes_to_channels_from_breaches(
        self, nodes: Nodes, potential_breaches: PotentialBreaches
    ) -> None:
        """Replace the content_pk of connection nodes using the nodes.breach_ids.

        This is required for exchange line assignment.

        Requires: content_type, line[:, 1]
        Sets: content_pk, content_type
        """
        is_connection_node = np.where(
            self.content_type == ContentType.TYPE_V2_CONNECTION_NODES
        )[0]
        node_ids = self.line[is_connection_node, 1]
        breach_id = nodes.breach_ids[nodes.id_to_index(node_ids), 0]
        has_breach = breach_id != -9999
        channel_id = potential_breaches.channel_id[
            potential_breaches.id_to_index(breach_id[has_breach])
        ]
        self.content_pk[is_connection_node[has_breach]] = channel_id
        self.content_type[is_connection_node[has_breach]] = ContentType.TYPE_V2_CHANNEL

    def assign_connection_nodes_to_channels_from_lines(
        self, nodes: Nodes, lines: Lines
    ):
        """Replace the content_pk of connection nodes on connected objects.

        The connection node will be assigned to the first 'connected' channel
        that is attached to it. Double connected channels get priority.

        This is required for exchange line assignment.

        Requires: content_type, line[:, 1]
        Sets: content_pk, content_type
        """
        PRIORITY = {LineType.LINE_1D_DOUBLE_CONNECTED: 0, LineType.LINE_1D_CONNECTED: 1}

        is_connection_node = np.where(
            self.content_type == ContentType.TYPE_V2_CONNECTION_NODES
        )[0]
        node_ids = self.line[is_connection_node, 1]

        line_halfs = LineHalfs.for_connection_nodes(
            nodes,
            lines,
            line_mask=(
                (lines.content_type == ContentType.TYPE_V2_CHANNEL)
                & (np.isin(lines.kcu, list(PRIORITY.keys())))
            ),
            node_mask=np.isin(nodes.id, node_ids),
        )
        # Order the line_halfs by (channel_id, calc type) and pick the first
        line_halfs.reorder(
            np.lexsort(
                [
                    line_halfs.content_pk,
                    replace(line_halfs.kcu, PRIORITY),
                    line_halfs.node_id,
                ]
            )
        )
        channel_id_per_node = line_halfs.first(line_halfs.content_pk)
        # Use the node_id -> channel_id map to set channel_id on self
        idx = search(
            line_halfs.get_reduce_id(),
            node_ids,
            check_exists=False,
            assume_ordered=True,
        )
        mask = idx != -9999

        self.content_pk[is_connection_node[mask]] = channel_id_per_node[idx[mask]]
        self.content_type[is_connection_node[mask]] = ContentType.TYPE_V2_CHANNEL

    def assign_exchange_lines(self, exchange_lines: ExchangeLines) -> None:
        """Assign exchange lines to the 1D-2D lines.

        Resets all content_type / content_pk.

        Requires: content_pk, content_type
        Sets: content_pk, content_type
        """
        line_idx_1, line_idx_2 = self.split_in_two(self.line[:, 1])
        self._assign_exchange_lines(line_idx_1, exchange_lines, is_primary=True)
        self._assign_exchange_lines(line_idx_2, exchange_lines, is_primary=False)

        # reset content_pk / content_type where not set to exchange line
        has_no_exc = self.content_type != ContentType.TYPE_V2_EXCHANGE_LINE
        self.content_type[has_no_exc] = -9999
        self.content_pk[has_no_exc] = -9999

    def _assign_exchange_lines(
        self, idx, exchange_lines: ExchangeLines, is_primary: bool
    ) -> None:
        if len(idx) == 0:
            return
        is_channel = self.content_type[idx] == ContentType.TYPE_V2_CHANNEL
        channel_id = self.content_pk[idx[is_channel]]
        content_pk = exchange_lines.get_for_channel_id(
            channel_id, is_primary=is_primary
        )
        self.content_pk[idx[is_channel]] = content_pk
        self.content_type[idx[is_channel]] = np.where(
            content_pk != -9999, ContentType.TYPE_V2_EXCHANGE_LINE, -9999
        )

    def get_1d_node_idx(self, nodes: Nodes):
        """Get the 1D node index based on line[:, 1]"""
        return nodes.id_to_index(self.line[:, 1])

    def assign_line_coords(self, nodes: Nodes):
        """Copy the 1D node location to the 2D side for non-exchange line lines.

        Returns an array of shapely Point geometries
        Requires: line[:, 1], content_pk, content_type
        Sets: line_coords (side_1d and side_2d)
        """
        self.line_coords[:, 2:] = nodes.coordinates[self.get_1d_node_idx(nodes)]
        self.line_coords[:, :2] = self.line_coords[:, 2:]

    def assign_2d_side_from_exchange_lines(self, exchange_lines: ExchangeLines):
        """Compute the closest Point on exchange line for each 1D2D line

        Requires: line_coords[:, 2:] (side_1d), content_pk, content_type
        Sets: line_coords[:, :2] (side_2d)
        """
        # Create an array of node indices & corresponding exchange line indices
        has_exc = self.content_type == ContentType.TYPE_V2_EXCHANGE_LINE

        # Find the closest points on the exchange lines
        exc_geoms = exchange_lines.the_geom[
            exchange_lines.id_to_index(self.content_pk[has_exc])
        ]
        shapely.prepare(exc_geoms)
        line_geom = shapely.shortest_line(exc_geoms, self.side_1d[has_exc])

        self.line_coords[has_exc, :2] = shapely.get_coordinates(
            shapely.get_point(line_geom, 0)
        )

    def assign_breaches(self, nodes: Nodes, potential_breaches: PotentialBreaches):
        """Assign breaches to the 1D-2D lines.

        It is assumed that the breach ids match the calculation type of the nodes.
        This is done in PotentialBreachPoints.match_breach_ids_with_calculation_types.

        If a node is double connected, the 1 or 2 breaches that are present need to be
        matched against existing lines. The matching minimizes the total distance of
        the 2D ends of the potential breach to the 2D end of the 1D-2D line (which
        is possibly derived from an exchange line).

        Requires: line_coords[:, :2]  (from the exchange line or 1D node location)
        Sets: line_coords[:, :2], line_geometries, content_pk, content_type
        """
        node_idx = self.get_1d_node_idx(nodes)
        breach_counts = np.count_nonzero(nodes.breach_ids[node_idx, :] != -9999, axis=1)
        line_idx_with_breaches = np.where(breach_counts > 0)[0]

        if len(line_idx_with_breaches) == 0:
            return

        # define a function that computes the distance between breach and line 2D sides
        breach_2d_side = potential_breaches.side_2d
        line_2d_side = self.side_2d

        def dist(breach_idx, line_idx):
            return shapely.distance(breach_2d_side[breach_idx], line_2d_side[line_idx])

        # a mapping with a breach_id for each line
        line_to_breach = {}

        # loop over lines with breaches (i is a line index)
        for i in np.where(breach_counts > 0)[0]:
            if i in line_to_breach:
                continue  # happens for double connected nodes

            # find all lines that are connected to the node
            line_idx = np.where(self.line[:, 1] == self.line[i, 1])[0]
            if len(line_idx) == 1:
                l1, l2 = line_idx[0], -9999
            else:
                l1, l2 = line_idx
            b1, b2 = potential_breaches.id_to_index(nodes.breach_ids[node_idx[i], :])

            if l2 == -9999 and b2 == -9999:
                # 1 line, 1 breach
                line_to_breach[l1] = b1
            elif l2 == -9999 and b2 != -9999:
                raise ValueError(
                    f"Node {nodes.id[node_idx[i]]} has two breaches assigned while it is not double connected"
                )
            elif b2 == -9999:
                # 2 lines, 1 breach; find closest line
                if dist(b1, l1) <= dist(b1, l2):
                    line_to_breach[l1] = b1
                else:
                    line_to_breach[l2] = b1
            elif b2 != -9999:
                # 2 lines, 2 breach; find the combination that minimizes total distance
                if (dist(b1, l1) + dist(b2, l2)) <= (dist(b1, l2) + dist(b2, l1)):
                    line_to_breach[l1], line_to_breach[l2] = b1, b2
                else:
                    line_to_breach[l1], line_to_breach[l2] = b2, b1

        line_idx, breach_idx = np.array(list(line_to_breach.items()), dtype=np.int32).T

        self.line_coords[line_idx, :2] = shapely.get_coordinates(
            breach_2d_side[breach_idx]
        )
        self.line_geometries[line_idx] = potential_breaches.the_geom[breach_idx]
        self.content_type[line_idx] = ContentType.TYPE_V2_BREACH
        self.content_pk[line_idx] = potential_breaches.id[breach_idx]

    def assign_2d_node(self, cell_tree: shapely.STRtree) -> None:
        """Assigns the 2D node id based on the line_coords

        Requires: line_coords[:, :2] (the 2D side)
        Sets: line[:, 0], line_coords[:] (clears)
        """
        # The query returns 2 1D arrays: one with indices into the supplied node
        # geometries and one with indices into the tree of cells.
        idx = cell_tree.query(self.side_2d)
        # Address edge cases of multiple 1D-2D lines per node: just take the one
        _, unique_matches = np.unique(idx[0], return_index=True)
        line_idx, cell_idx = idx[:, unique_matches]
        self.line[line_idx, 0] = cell_idx
        self.line_coords[:] = np.nan

    def check_unassigned(self, nodes) -> None:
        """Checks 1D-2D lines where any of the required nodes is set to null, represented as -9999
        This is the case when the nodes are outside the 2D domain.
        """
        invalid_rows = self.line[:, 0] == -9999
        if invalid_rows.any():
            invalid_node_ids = self.line[invalid_rows, 1]
            invalid_nodes = nodes.id_to_index(invalid_node_ids)
            invalid_nodes_formatted = nodes.format_message(invalid_nodes)

            raise SchematisationError(
                f"The following objects are connected but are (partially) outside of the 2D model domain: {invalid_nodes_formatted}."
            )

    def transfer_2d_node_to_groundwater(self, offset: int):
        """Transfers the 1D-2D line to a groundwater node

        Sets: line[:, 0]
        """
        has_2d_node = np.where(self.line[:, 0] != -9999)[0]
        if offset == 0:
            # no groundwater cells; reset the node id (will raise error later)
            self.line[has_2d_node, 0] = -9999
        else:
            self.line[has_2d_node, 0] += offset

    def assign_kcu(self, mask, is_closed) -> None:
        """Set kcu where it is not set already"""
        node_id = self.line[mask, 1]
        node_id_unique, counts = np.unique(node_id, return_counts=True)
        line_is_double = np.isin(self.line[mask, 1], node_id_unique[counts == 2])
        self.kcu[mask] = np.choose(
            line_is_double * 2 + is_closed,
            choices=[
                LineType.LINE_1D2D_SINGLE_CONNECTED_OPEN_WATER,
                LineType.LINE_1D2D_SINGLE_CONNECTED_CLOSED,
                LineType.LINE_1D2D_DOUBLE_CONNECTED_OPEN_WATER,
                LineType.LINE_1D2D_DOUBLE_CONNECTED_CLOSED,
            ],
        )

    def assign_dpumax_from_breaches(self, potential_breaches: PotentialBreaches):
        """Set the dpumax based exchange_lines.exchange_level"""
        has_breach = self.content_type == ContentType.TYPE_V2_BREACH
        self.assign_dpumax(
            has_breach,
            potential_breaches.exchange_level[
                potential_breaches.id_to_index(self.content_pk[has_breach])
            ],
        )

    def assign_dpumax_from_exchange_lines(self, exchange_lines: ExchangeLines):
        """Set the dpumax based exchange_lines.exchange_level"""
        has_exc = self.content_type == ContentType.TYPE_V2_EXCHANGE_LINE
        self.assign_dpumax(
            has_exc,
            exchange_lines.exchange_level[
                exchange_lines.id_to_index(self.content_pk[has_exc])
            ],
        )

    def assign_dpumax_from_obstacles(self, obstacles: Obstacles):
        """Set the dpumax based on intersected obstacles

        Only for open connections. Requires line_geometries.
        """
        is_open_water = np.isin(
            self.kcu,
            [
                LineType.LINE_1D2D_SINGLE_CONNECTED_OPEN_WATER,
                LineType.LINE_1D2D_DOUBLE_CONNECTED_OPEN_WATER,
            ],
        )
        is_open_water_idx = np.where(is_open_water)[0]
        obstacle_crest_levels, obstacle_idx = obstacles.compute_dpumax(
            self, where=is_open_water_idx
        )
        self.assign_dpumax(is_open_water, obstacle_crest_levels)
        mask = obstacle_idx != -9999
        self.assign_ds1d_half_from_obstacles(
            obstacles, is_open_water_idx[mask], obstacle_idx[mask]
        )

    def assign_ds1d_half_from_obstacles(
        self, obstacles: Obstacles, line_idx, obstacle_idx
    ):
        if len(line_idx) == 0:
            return
        line_geoms = self.line_geometries[line_idx]
        # compute the intersections (use shortest_line and not intersects to
        # account for the case that the obstacle and the line intersect multiple times)
        points = shapely.get_point(
            shapely.shortest_line(obstacles.the_geom[obstacle_idx], line_geoms), 0
        )
        # compute the distance to the start of the line
        self.ds1d_half[line_idx] = shapely.line_locate_point(line_geoms, points)

    def assign_dpumax(self, mask, dpumax) -> None:
        """Assign dpumax only where it is not set already"""
        is_nan = np.isnan(self.dpumax)
        self.dpumax[mask & is_nan] = dpumax[is_nan[mask]]

    def assign_ds1d(self, nodes: Nodes) -> None:
        """Sets the length (ds1d) based on the 2D cell with, where not yet set.

        Requires: line[:, 0]
        Sets: ds1d
        """
        has_no_ds1d = np.isnan(self.ds1d)
        cell_idx = nodes.id_to_index(self.line[has_no_ds1d, 0])
        self.ds1d[has_no_ds1d] = nodes.bounds[cell_idx, 2] - nodes.bounds[cell_idx, 0]

    def assign_ds1d_half(self) -> None:
        """Sets the velocity point location on the line (ds1d_half), where not yet set.

        Requires: ds1d
        Sets: ds1d_half
        """
        has_no_ds1d_half = np.isnan(self.ds1d_half)
        self.ds1d_half[has_no_ds1d_half] = (
            shapely.length(self.line_geometries[has_no_ds1d_half]) / 2
        )

    def output_breaches(self, breaches: PotentialBreaches) -> PotentialBreaches:
        """Create the actual breaches based on the 1D-2D lines

        This results in 'named' breaches (which are present in the input potential_breaches).
        Other 1D-2D lines may be breached as well, but they have no metadata.
        """
        line_idx = np.where(self.content_type == ContentType.TYPE_V2_BREACH)[0]
        breach_idx = breaches.id_to_index(self.content_pk[line_idx])

        # Order by breach id
        sorter = np.argsort(breach_idx)
        line_idx = line_idx[sorter]
        breach_idx = breach_idx[sorter]

        # Filter by breaches that miss metadata
        mask = (breaches.levee_material[breach_idx] != -9999) & np.isfinite(
            breaches.maximum_breach_depth[breach_idx]
        )
        line_idx = line_idx[mask]
        breach_idx = breach_idx[mask]

        return PotentialBreaches(
            id=range(len(breach_idx)),
            line_id=self.id[line_idx],
            content_pk=breaches.id[breach_idx],
            maximum_breach_depth=breaches.maximum_breach_depth[breach_idx],
            levee_material=breaches.levee_material[breach_idx],
            the_geom=self.get_velocity_points(line_idx),
            code=breaches.code[breach_idx],
            display_name=breaches.display_name[breach_idx],
        )

    @classmethod
    def create_groundwater(
        cls, nodes: Nodes, line_id_counter: Iterator[int]
    ) -> "Lines1D2D":
        """Create a 1D-2D groundwater lines for each 1D node.

        Sets: id, line[:, 1], kcu
        """
        node_idx = np.where(nodes.has_groundwater_exchange)[0]
        line = np.full((len(node_idx), 2), -9999, dtype=np.int32)
        line[:, 1] = nodes.index_to_id(node_idx)
        return cls(
            id=itertools.islice(line_id_counter, len(node_idx)),
            line=line,
            kcu=LineType.LINE_1D2D_GROUNDWATER,
        )

    def assign_groundwater_exchange(self, nodes: Nodes, cn: ConnectionNodes):
        """Compute the hydraulic_conductivity values for groundwater exchange manholes
        when present. Set all dpumax values for 1d2d groundwater exchange lines.

        Sets:
        - hydraulic_conductivity_in for manholes
        - hydraulic_conductivity_out for manholes
        - dpumax (from 1D nodes dmax)
        """

        node_idx = self.get_1d_node_idx(nodes)
        idx = np.where(
            nodes.content_type[node_idx] == ContentType.TYPE_V2_CONNECTION_NODES
        )[0]
        cn_idx = cn.id_to_index(nodes.content_pk[node_idx[idx]])

        mask = cn.has_groundwater_exchange[cn_idx]
        idx = idx[mask]
        cn_idx = cn_idx[mask]

        self.hydraulic_resistance_in[idx] = (
            cn.hydraulic_conductivity_in[cn_idx] / cn.exchange_thickness[cn_idx]
        )
        self.hydraulic_resistance_out[idx] = (
            cn.hydraulic_conductivity_out[cn_idx] / cn.exchange_thickness[cn_idx]
        )

        self.dpumax = nodes[self.get_1d_node_idx(nodes)].dmax



import logging

import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_equal

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.constants import CalculationType, ContentType
from threedigrid_builder.exceptions import SchematisationError
from threedigrid_builder.grid import Channels, PotentialBreaches, PotentialBreachPoints


def test_potential_breach_sides():
    potential_breaches = PotentialBreaches(
        id=[1],
        channel_id=[1],
        the_geom=[shapely.linestrings([[0, 0], [0, 1]])],
    )
    assert_almost_equal(shapely.get_coordinates(potential_breaches.side_1d), [[0, 0]])
    assert_almost_equal(shapely.get_coordinates(potential_breaches.side_2d), [[0, 1]])


def test_potential_breach_merge(caplog):
    breaches = PotentialBreachPoints(
        linestrings=Channels(
            id=[0, 1],
            the_geom=shapely.linestrings([[[0, 0], [10, 0]], [[0, 0], [0, 10]]]),
        ).linestrings,
        id=range(7),
        s1d=[0.00001, 2.0, 5.0, 5.00001, 9.99999, 4, 4.00001],
        linestring_idx=[0, 0, 0, 0, 0, 1, 1],
        content_pk=range(1, 8),
    )

    with caplog.at_level(logging.WARNING):
        actual = breaches.merge()

    assert len(caplog.messages) == 0

    assert_almost_equal(actual.s1d, [0.0, 2.0, 5.0, 10.0, 4.0])
    assert_almost_equal(actual.linestring_idx, [0, 0, 0, 0, 1])
    assert_almost_equal(actual.content_pk, [1, 2, 3, 5, 6])
    assert_almost_equal(actual.secondary_content_pk, [-9999, -9999, 4, -9999, 7])


def test_potential_breach_merge_empty():
    actual = PotentialBreachPoints.empty(
        linestrings=Channels(
            id=[0, 1],
            the_geom=shapely.linestrings([[[0, 0], [10, 0]], [[0, 0], [0, 10]]]),
        ).linestrings
    ).merge()

    assert len(actual) == 0


@pytest.fixture
def threeway_junction():
    nodes = Nodes(
        id=range(1, 5),
        content_type=ContentType.TYPE_V2_CONNECTION_NODES,
        content_pk=[1, 2, 3, 4],
    )
    linestrings = Channels(
        id=[11, 12, 13],
        the_geom=shapely.linestrings([[[0, 0], [0, 1]]] * 3),
    ).linestrings
    lines = Lines(
        id=range(3),
        line=[(1, 2), (1, 3), (4, 1)],
        content_type=ContentType.TYPE_V2_CHANNEL,
        content_pk=[11, 12, 13],
    )
    return nodes, lines, linestrings


@pytest.mark.parametrize(
    "breach_points,expected",
    [
        (  # no breach points: nothing assigned
            {"id_1": [], "id_2": [], "s1d": [], "ch_idx": []},
            (-9999, -9999),
        ),
        (  # breach point at start, single connected
            {"id_1": [5], "id_2": [-9999], "s1d": [0.0], "ch_idx": [0]},
            (5, -9999),
        ),
        (  # breach point at end, single connected (linestring 2 is swapped)
            {"id_1": [7], "id_2": [-9999], "s1d": [1.0], "ch_idx": [2]},
            (7, -9999),
        ),
        (  # breach point at start, far side
            {"id_1": [5], "id_2": [-9999], "s1d": [1.0], "ch_idx": [0]},
            (-9999, -9999),
        ),
        (  # breach point at end, far side (linestring 2 is swapped)
            {"id_1": [7], "id_2": [-9999], "s1d": [0.0], "ch_idx": [2]},
            (-9999, -9999),
        ),
        (  # breach point at start, double connected
            {"id_1": [5], "id_2": [6], "s1d": [0.0], "ch_idx": [0]},
            (5, 6),
        ),
        (  # breach point at end, double connected (linestring 2 is swapped)
            {"id_1": [7], "id_2": [8], "s1d": [1.0], "ch_idx": [2]},
            (7, 8),
        ),
        (  # two breach points, single connected
            {"id_1": [5, 6], "id_2": -9999, "s1d": [0.0, 0.0], "ch_idx": [0, 1]},
            (5, -9999),
        ),
        (  # two breach points, single connected (orders by ch_idx)
            {"id_1": [5, 6], "id_2": -9999, "s1d": [0.0, 0.0], "ch_idx": [1, 0]},
            (6, -9999),
        ),
        (  # two breach points, one double connected (gets priority)
            {"id_1": [5, 6], "id_2": [-9999, 8], "s1d": [0.0, 0.0], "ch_idx": [0, 1]},
            (6, 8),
        ),
    ],
)
def test_assign_to_connection_nodes(threeway_junction, breach_points, expected):
    """Set the breach ids on a connection node with 3 channels"""
    nodes, lines, linestrings = threeway_junction
    breach_points = PotentialBreachPoints(
        linestrings=linestrings,
        id=range(len(breach_points["id_1"])),
        content_pk=breach_points["id_1"],
        secondary_content_pk=breach_points["id_2"],
        s1d=breach_points["s1d"],
        linestring_idx=breach_points["ch_idx"],
    )
    breach_points.assign_to_connection_nodes(nodes, lines)
    assert_equal(nodes.breach_ids[0], expected)


@pytest.mark.parametrize(
    "calculation_type,breach_ids",
    [
        (CalculationType.ISOLATED, [-9999, -9999]),
        (CalculationType.CONNECTED, [1, -9999]),
        (CalculationType.DOUBLE_CONNECTED, [1, 2]),
    ],
)
def test_match_breach_ids_with_calculation_types_ok(calculation_type, breach_ids):
    nodes = Nodes(
        id=[1],
        calculation_type=calculation_type,
        breach_ids=[breach_ids],
        content_type=ContentType.TYPE_V2_CHANNEL,
        content_pk=[11],
    )
    nodes.breach_ids[0, :] = breach_ids

    PotentialBreachPoints.match_breach_ids_with_calculation_types(nodes)


@pytest.mark.parametrize(
    "calculation_type,breach_ids,msg",
    [
        (
            CalculationType.ISOLATED,
            [1, -9999],
            "The following objects have potential breaches, but are not (double) connected: channels [11].",
        ),
        (
            CalculationType.ISOLATED,
            [1, 2],
            "The following objects have potential breaches, but are not (double) connected: channels [11].",
        ),
        (
            CalculationType.CONNECTED,
            [1, 2],
            "The following objects have two potential breaches at the same position, but are not double connected: channels [11].",
        ),
    ],
)
def test_match_breach_ids_with_calculation_types_err(calculation_type, breach_ids, msg):
    nodes = Nodes(
        id=[1],
        calculation_type=calculation_type,
        breach_ids=[breach_ids],
        content_type=ContentType.TYPE_V2_CHANNEL,
        content_pk=[11],
    )
    nodes.breach_ids[0, :] = breach_ids

    with pytest.raises(SchematisationError) as e:
        PotentialBreachPoints.match_breach_ids_with_calculation_types(nodes)

    assert str(e.value) == msg



import itertools
import re
from contextlib import nullcontext as does_not_raise
from unittest import mock

import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_array_equal
from shapely.testing import assert_geometries_equal

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.constants import CalculationType, ContentType, LineType
from threedigrid_builder.exceptions import SchematisationError
from threedigrid_builder.grid import (
    ConnectionNodes,
    ExchangeLines,
    Lines1D2D,
    Obstacles,
    PotentialBreaches,
)

ISO = CalculationType.ISOLATED
C1 = CalculationType.CONNECTED
C2 = CalculationType.DOUBLE_CONNECTED
CN = ContentType.TYPE_V2_CONNECTION_NODES
CH = ContentType.TYPE_V2_CHANNEL
PIPE = ContentType.TYPE_V2_PIPE
EXC = ContentType.TYPE_V2_EXCHANGE_LINE
BREACH = ContentType.TYPE_V2_BREACH


@pytest.mark.parametrize(
    "calculation_type,expected_line",
    [
        ([], []),
        ([ISO], []),
        ([C1], [1]),
        ([C2], [1, 1]),
        ([C1, ISO, C2], [1, 3, 3]),
        ([C2, ISO, C1], [1, 1, 3]),
    ],
)
def test_create_lines_1d2d(calculation_type, expected_line):
    nodes = Nodes(
        id=range(1, len(calculation_type) + 1),
        calculation_type=calculation_type,
        content_pk=3,
        content_type=ContentType.TYPE_V2_CHANNEL,
    )
    actual = Lines1D2D.create(nodes, itertools.count())
    assert_array_equal(actual.line[:, 0], -9999)
    assert_array_equal(actual.line[:, 1], expected_line)
    assert_array_equal(actual.content_pk, 3)
    assert_array_equal(actual.content_type, ContentType.TYPE_V2_CHANNEL)


@pytest.mark.parametrize(
    "breach_ids,content_type,expected_content_pk,expected_content_type",
    [
        ((-9999, -9999), CN, 33, CN),
        ((5, -9999), CN, 11, CH),
        ((5, 6), CN, 11, CH),
        ((5, 6), CH, 33, CH),
    ],
)
def test_assign_connection_nodes_to_channels_from_breaches(
    breach_ids, content_type, expected_content_pk, expected_content_type
):
    nodes = Nodes(id=[1], breach_ids=[breach_ids])
    lines = Lines1D2D(
        id=[1], content_type=[content_type], content_pk=[33], line=[(-9999, 1)]
    )
    potential_breaches = PotentialBreaches(id=[5, 6], channel_id=[11, 11])
    lines.assign_connection_nodes_to_channels_from_breaches(nodes, potential_breaches)

    assert_array_equal(lines.content_pk, [expected_content_pk])
    assert_array_equal(lines.content_type, [expected_content_type])


@pytest.mark.parametrize(
    "content_pk,content_type,exc_line_channel_id,expected_content_pk,expected_content_type",
    [
        ([], CH, [], [], []),
        ([11], CH, [], [-9999], [-9999]),
        ([12], CH, [11], [-9999], [-9999]),
        ([12], CH, [12], [1], [EXC]),
        ([12, 12], CH, [12], [1, -9999], [EXC, -9999]),
        ([12, 12], CH, [12, 11], [1, -9999], [EXC, -9999]),
        ([12, 12], CH, [12, 12], [1, 2], [EXC, EXC]),
        ([12], CN, [12], [-9999], [-9999]),
    ],
)
def test_assign_exchange_lines(
    content_pk,
    content_type,
    exc_line_channel_id,
    expected_content_pk,
    expected_content_type,
):
    lines = Lines1D2D(
        id=range(len(content_pk)),
        content_pk=content_pk,
        content_type=content_type,
    )
    exchange_lines = ExchangeLines(
        id=range(1, len(exc_line_channel_id) + 1), channel_id=exc_line_channel_id
    )
    lines.assign_exchange_lines(exchange_lines)

    assert_array_equal(lines.content_pk, expected_content_pk)
    assert_array_equal(lines.content_type, expected_content_type)


def test_assign_line_coords():
    nodes = Nodes(
        id=[1, 2, 3],
        coordinates=[(2, 5), (4, 5), (6, 5)],
    )
    lines = Lines1D2D(id=range(2), line=[[-9999, 1], [-9999, 3]])
    lines.assign_line_coords(nodes)

    assert_array_equal(lines.line_coords, [[2, 5, 2, 5], [6, 5, 6, 5]])


def test_assign_2d_side_from_exchange_lines():
    exchange_lines = ExchangeLines(
        id=[1, 2],
        the_geom=shapely.linestrings([[[0, 0], [10, 0]], [[0, 10], [10, 10]]]),
    )
    lines = Lines1D2D(
        id=range(4),
        line=[[-9999, 1], [-9999, 1], [-9999, 2], [-9999, 3]],
        line_coords=[[2, 5, 2, 5], [2, 5, 2, 5], [4, 5, 4, 5], [6, 5, 6, 5]],
        content_pk=[1, 2, 1, -9999],
        content_type=[EXC, EXC, EXC, -9999],
    )
    lines.assign_2d_side_from_exchange_lines(exchange_lines)

    assert_array_equal(lines.line_coords[:, :2], [[2, 0], [2, 10], [4, 0], [6, 5]])


@pytest.fixture
def cell_tree():
    return shapely.STRtree([shapely.box(0, 0, 1, 1), shapely.box(1, 0, 3, 2)])


@pytest.mark.parametrize(
    "side_2d_coordinates,expected_2d_node_id",
    [
        ([(0.5, 0.5)], [0]),  # first cell, center
        ([(0, 0.5)], [0]),  # first cell, left edge
        ([(0.5, 1)], [0]),  # first cell, top edge
        ([(0.5, 0)], [0]),  # first cell, bottom edge
        ([(0, 1)], [0]),  # first cell, topleft corner
        ([(0, 0)], [0]),  # first cell, bottomleft corner
        ([(2, 1)], [1]),  # second cell, center
        ([(3, 1)], [1]),  # second cell, right edge
        ([(2, 2)], [1]),  # second cell, top edge
        ([(2, 0)], [1]),  # second cell, bottom edge
        ([(3, 2)], [1]),  # second cell, topright corner
        ([(3, 0)], [1]),  # second cell, bottomright corner
        ([(1, 1)], [0]),  # edge between: top corner
        ([(1, 0)], [0]),  # edge between: bottom corner
        ([(1, 0.5)], [0]),  # edge between: middle
        ([(0.5, 0.5), (0.5, 0.9)], [0, 0]),  # two cells, same
        ([(0.5, 0.5), (2, 1)], [0, 1]),  # two cells, different
        ([(3, 3)], [-9999]),  # out of bounds
    ],
)
def test_assign_2d_node(side_2d_coordinates, expected_2d_node_id, cell_tree):
    lines = Lines1D2D(
        id=range(len(side_2d_coordinates)),
        line_coords=[x + (np.nan, np.nan) for x in side_2d_coordinates],
    )

    lines.assign_2d_node(cell_tree)

    assert_array_equal(lines.line[:, 0], expected_2d_node_id)
    assert_array_equal(lines.line_coords, np.nan)  # is cleared


@pytest.mark.parametrize(
    "nodes,lines,expectation",
    [
        (
            Nodes(
                id=[1, 2], content_type=[12, 12], content_pk=[5, 6], node_type=[4, 4]
            ),
            Lines1D2D(
                id=[3, 4],
                line=[(-9999, 1), (1, 2)],
                line_coords=[(10, 0, 0, 0), (5, 2, 3, 8)],
            ),
            pytest.raises(
                SchematisationError,
                match=re.escape(
                    "The following objects are connected but are (partially) outside of the 2D model domain: connection nodes [5]."
                ),
            ),
        ),
        (
            Nodes(
                id=[1, 2], content_type=[12, 12], content_pk=[5, 6], node_type=[4, 4]
            ),
            Lines1D2D(
                id=[3, 4],
                line=[(2, 1), (1, 2)],
                line_coords=[(10, 0, 0, 0), (5, 2, 3, 8)],
            ),
            does_not_raise(),
        ),
    ],
)
def test_check_line_unassigned_nodes(nodes, lines, expectation):
    with expectation:
        lines.check_unassigned(nodes)


def test_assign_kcu():
    lines = Lines1D2D(id=range(7), line=[[-9999] + [x] for x in [1, 1, 2, 2, 3, 4, 5]])
    lines.content_type[0] = ContentType.TYPE_V2_BREACH

    lines.assign_kcu(
        mask=np.array([1, 1, 1, 1, 0, 1, 1], dtype=bool),
        is_closed=np.array([0, 0, 1, 1, 0, 1], dtype=bool),
    )

    assert_array_equal(
        lines.kcu,
        [
            LineType.LINE_1D2D_DOUBLE_CONNECTED_OPEN_WATER,
            LineType.LINE_1D2D_DOUBLE_CONNECTED_OPEN_WATER,
            LineType.LINE_1D2D_DOUBLE_CONNECTED_CLOSED,
            LineType.LINE_1D2D_DOUBLE_CONNECTED_CLOSED,
            -9999,
            LineType.LINE_1D2D_SINGLE_CONNECTED_OPEN_WATER,
            LineType.LINE_1D2D_SINGLE_CONNECTED_CLOSED,
        ],
    )


@pytest.mark.parametrize(
    "existing,mask,expected",
    [
        ([np.nan, np.nan, np.nan], [1, 0, 1], [3.0, np.nan, 6.0]),
        ([5.0, 5.0, 5.0], [1, 0, 1], [5.0, 5.0, 5.0]),
        ([5.0, np.nan, np.nan], [1, 0, 1], [5.0, np.nan, 6.0]),
    ],
)
def test_assign_dpumax(existing, mask, expected):
    lines = Lines1D2D(id=range(3), dpumax=existing)

    lines.assign_dpumax(
        np.array(mask, dtype=bool),
        np.array([3.0, 6.0]),
    )

    assert_array_equal(lines.dpumax, expected)


def test_get_1d_node_idx():
    nodes = Nodes(
        id=[1, 2, 3],
        coordinates=[(2, 5), (4, 5), (6, 5)],
    )
    lines = Lines1D2D(
        id=range(4),
        line=[[-9999, 1], [-9999, 1], [-9999, 2], [-9999, 3]],
    )
    actual = lines.get_1d_node_idx(nodes)

    assert_array_equal(actual, [0, 0, 1, 2])


@mock.patch.object(Lines1D2D, "assign_dpumax")
def test_assign_dpumax_from_exchange_lines(assign_dpumax):
    lines = Lines1D2D(id=range(3), content_pk=[1, 2, 3], content_type=[EXC, -9999, EXC])
    exchange_lines = ExchangeLines(id=[1, 2, 3], exchange_level=[1.2, 2.3, np.nan])

    lines.assign_dpumax_from_exchange_lines(exchange_lines)

    (actual_mask, actual_dpumax), _ = assign_dpumax.call_args

    assert_array_equal(actual_mask, [1, 0, 1])
    assert_array_equal(actual_dpumax, [1.2, np.nan])


@mock.patch.object(Lines1D2D, "assign_dpumax")
def test_assign_dpumax_from_breaches(assign_dpumax):
    lines = Lines1D2D(
        id=range(3), content_pk=[1, 2, 3], content_type=[BREACH, -9999, BREACH]
    )
    breaches = PotentialBreaches(id=[1, 2, 3], exchange_level=[1.2, 2.3, np.nan])

    lines.assign_dpumax_from_breaches(breaches)

    (actual_mask, actual_dpumax), _ = assign_dpumax.call_args

    assert_array_equal(actual_mask, [1, 0, 1])
    assert_array_equal(actual_dpumax, [1.2, np.nan])


@mock.patch.object(Lines1D2D, "assign_dpumax")
@mock.patch.object(Lines1D2D, "assign_ds1d_half_from_obstacles")
def test_assign_dpumax_from_obstacles(
    assign_ds1d_half_from_obstacles,
    assign_dpumax,
):
    obstacles = mock.Mock()
    obstacles.compute_dpumax.return_value = (
        np.array([1.2, np.nan]),
        np.array([1, -9999]),
    )

    lines = Lines1D2D(
        id=range(3),
        kcu=[
            LineType.LINE_1D2D_SINGLE_CONNECTED_OPEN_WATER,
            LineType.LINE_1D2D_SINGLE_CONNECTED_CLOSED,
            LineType.LINE_1D2D_SINGLE_CONNECTED_OPEN_WATER,
        ],
    )

    lines.assign_dpumax_from_obstacles(obstacles)

    args, kwargs = obstacles.compute_dpumax.call_args
    assert len(args) == 1
    assert args[0] is lines
    assert len(kwargs) == 1
    assert_array_equal(kwargs["where"], [0, 2])

    (actual_mask, actual_dpumax), _ = assign_dpumax.call_args

    assert_array_equal(actual_mask, [1, 0, 1])
    assert_array_equal(actual_dpumax, [1.2, np.nan])

    args, _ = assign_ds1d_half_from_obstacles.call_args

    assert args[0] is obstacles
    assert_array_equal(args[1], [0])
    assert_array_equal(args[2], [1])


@pytest.mark.parametrize(
    "breach_ids,breach_2d_coords,content_pk",
    [
        ([-9999, -9999], [(10, 0)], [-9999]),
        ([1, -9999], [(10, 0)], [1]),
    ],
)
def test_assign_breaches_single_connected(breach_ids, breach_2d_coords, content_pk):
    nodes = Nodes(id=[1], breach_ids=[breach_ids])
    lines = Lines1D2D(id=[1], line=[(-9999, 1)], line_coords=[(10, 0, 0, 0)])
    potential_breaches = PotentialBreaches(
        id=range(1, len(breach_2d_coords) + 1),
        the_geom=shapely.linestrings([[(0, 0), x] for x in breach_2d_coords]),
    )
    lines.assign_breaches(nodes, potential_breaches)

    assert_array_equal(
        lines.content_type, np.where(np.array(content_pk) != -9999, BREACH, -9999)
    )
    assert_array_equal(lines.content_pk, content_pk)


def test_assign_breaches_single_connected_err():
    nodes = Nodes(id=[1], breach_ids=[[1, 2]])
    lines = Lines1D2D(id=[1], line=[(-9999, 1)], line_coords=[(10, 0, 0, 0)])
    potential_breaches = PotentialBreaches(
        id=range(1, 3),
    )
    with pytest.raises(ValueError):
        lines.assign_breaches(nodes, potential_breaches)


@pytest.mark.parametrize(
    "breach_ids,breach_2d_coords,content_pk",
    [
        ([-9999, -9999], [(10, 0)], [-9999, -9999]),
        ([1, -9999], [(10, 0)], [1, -9999]),
        ([1, -9999], [(10, 10)], [-9999, 1]),
        ([1, -9999], [(10, 5)], [1, -9999]),
        ([1, 2], [(10, 0), (10, 10)], [1, 2]),
        ([1, 2], [(10, 10), (10, 0)], [2, 1]),
        ([1, 2], [(10, 1), (10, 9)], [1, 2]),
        ([1, 2], [(10, 9), (10, 1)], [2, 1]),
        ([1, 2], [(10, 5), (10, 5)], [1, 2]),
    ],
)
def test_assign_breaches_double_connected(breach_ids, breach_2d_coords, content_pk):
    nodes = Nodes(id=[1], breach_ids=[breach_ids])
    lines = Lines1D2D(
        id=[1, 2],
        line=[(-9999, 1), (-9999, 1)],
        line_coords=[(10, 0, 0, 0), (10, 10, 0, 0)],
    )
    potential_breaches = PotentialBreaches(
        id=range(1, len(breach_2d_coords) + 1),
        the_geom=shapely.linestrings([[(0, 0), x] for x in breach_2d_coords]),
    )
    lines.assign_breaches(nodes, potential_breaches)

    assert_array_equal(
        lines.content_type, np.where(np.array(content_pk) != -9999, BREACH, -9999)
    )
    assert_array_equal(lines.content_pk, content_pk)


def test_assign_breaches_multiple():
    nodes = Nodes(
        id=[1, 2, 3],
        coordinates=[(2, 5), (4, 5), (6, 5)],
        breach_ids=[(1, 2), (-9999, -9999), (4, -9999)],
    )
    lines = Lines1D2D(
        id=range(4),
        line=[[-9999, 1], [-9999, 1], [-9999, 2], [-9999, 3]],
        line_coords=[x + [0, 0] for x in [[2, 0], [2, 10], [4, 0], [6, 5]]],
        content_type=[EXC, EXC, EXC, -9999],
        content_pk=[0, 1, 2, 3],
    )
    potential_breaches = PotentialBreaches(
        id=[1, 2, 3, 4],
        the_geom=shapely.linestrings(
            [[(0, 0), x] for x in [(2, 9), (2, 1), (6, 10), (6, 1)]]
        ),
        code=["a", "b", "c", "d"],
    )
    lines.assign_breaches(nodes, potential_breaches)

    # also check the other fields that are set by assign_breaches
    assert_array_equal(lines.line_coords[:, :2], [(2, 1), (2, 9), (4, 0), (6, 1)])
    assert_array_equal(lines.content_type, [BREACH, BREACH, EXC, BREACH])
    assert_array_equal(lines.content_pk, [2, 1, 2, 4])


@pytest.mark.parametrize(
    "obstacle_geom,line_coords,expected",
    [
        ([[5, 0], [5, 10]], [0, 5, 10, 5], 5.0),
        ([[5, 0], [5, 10]], [2, 5, 10, 5], 3.0),
        ([[5, 0], [5, 10]], [0, 0, 10, 0], 5.0),
        ([[5, 1], [5, 10]], [0, 0, 10, 0], 5.0),
        ([[5, 3], [5, 10]], [5, 0, 5, 10], 3.0),
    ],
)
def test_assign_ds1d_half_from_obstacles(obstacle_geom, line_coords, expected):
    obstacles = Obstacles(id=[1], the_geom=[shapely.linestrings(obstacle_geom)])
    lines = Lines1D2D(id=range(1), line_coords=[line_coords])
    lines.fix_line_geometries()

    lines.assign_ds1d_half_from_obstacles(obstacles, np.array([0]), np.array(([0])))

    assert_almost_equal(lines.ds1d_half, [expected])


def test_assign_ds1d():
    nodes = Nodes(id=[1, 2, 3], bounds=[(0, 0, 8, 8), (8, 0, 10, 2), (8, 2, 10, 4)])
    lines = Lines1D2D(
        id=range(4),
        line=[[1, -9999], [1, -9999], [2, -9999], [3, -9999]],
        ds1d=[np.nan, np.nan, np.nan, 5.2],
    )
    lines.assign_ds1d(nodes)

    assert_array_equal(lines.ds1d, [8.0, 8.0, 2.0, 5.2])


def test_assign_ds1d_half():
    lines = Lines1D2D(
        id=range(2),
        ds1d_half=[np.nan, 5.0],
        line_geometries=[shapely.linestrings([[0, 0], [0, 9]]), None],
    )
    lines.assign_ds1d_half()

    assert_array_equal(lines.ds1d_half, [4.5, 5.0])


@pytest.fixture
def threeway_junction():
    nodes = Nodes(
        id=range(1, 5),
        content_type=ContentType.TYPE_V2_CONNECTION_NODES,
        content_pk=[1, 2, 3, 4],
    )
    lines = Lines(
        id=range(3),
        line=[(1, 2), (1, 3), (4, 1)],
        content_type=ContentType.TYPE_V2_CHANNEL,
        content_pk=[11, 12, 13],
        ds1d=[1.2, 2.3, 3.4],
        ds1d_half=[0.6, 1.0, 1.7],
    )
    return nodes, lines


@pytest.fixture
def potential_breaches():
    return PotentialBreaches(
        id=[5, 6, 7],
        channel_id=[11, 12, 13],
    )


@pytest.mark.parametrize(
    "kcu,expected",
    [
        ([ISO, ISO, ISO], 33),
        ([C1, ISO, ISO], 11),
        ([ISO, ISO, C1], 13),
        ([ISO, C1, C1], 12),
        ([C2, ISO, ISO], 11),
        ([ISO, ISO, C2], 13),
        ([ISO, C2, C2], 12),
        ([C1, C2, C2], 12),
    ],
)
@pytest.mark.parametrize("is_double_connected", [True, False])
def test_assign_connection_nodes_to_channels_from_lines(
    threeway_junction, kcu, expected, is_double_connected
):
    threeway_junction[1].kcu[:] = kcu
    n = 2 if is_double_connected else 1
    lines_1d2d = Lines1D2D(
        id=range(n), content_type=CN, content_pk=33, line=[(-9999, 1)] * n
    )
    lines_1d2d.assign_connection_nodes_to_channels_from_lines(*threeway_junction)

    assert_array_equal(lines_1d2d.content_pk, expected)
    assert_array_equal(lines_1d2d.content_type, CN if expected == 33 else CH)


@mock.patch.object(Lines1D2D, "get_velocity_points")
def test_output_breaches(get_velocity_points):
    lines = Lines1D2D(
        id=range(1, 5),
        content_type=[-9999, BREACH, BREACH, BREACH],
        content_pk=[-9999, 1, 2, 3],
    )
    potential_breaches = PotentialBreaches(
        id=[1, 2, 3],
        code=["a", "b", "c"],
        display_name=["aa", "bb", "cc"],
        levee_material=[1, 2, -9999],
        maximum_breach_depth=[np.nan, 1.3, 1.4],
    )

    get_velocity_points.return_value = shapely.points([[0, 0]])
    actual = lines.output_breaches(potential_breaches)

    assert_array_equal(actual.id, [0])
    assert_array_equal(actual.line_id, [3])
    assert_array_equal(actual.content_pk, [2])

    assert_array_equal(actual.maximum_breach_depth, [1.3])
    assert_array_equal(actual.levee_material, [2])

    assert_geometries_equal(actual.the_geom, get_velocity_points.return_value)

    assert_array_equal(actual.code, ["b"])
    assert_array_equal(actual.display_name, ["bb"])

    get_velocity_points.assert_called_once()


@pytest.mark.parametrize("has_groundwater_exchange", [True, False])
def test_create_lines_1d2d_groundwater(has_groundwater_exchange):
    nodes = Nodes(
        id=[1], content_pk=3, has_groundwater_exchange=has_groundwater_exchange
    )
    actual = Lines1D2D.create_groundwater(nodes, itertools.count())

    if has_groundwater_exchange:
        assert_array_equal(actual.line, [[-9999, 1]])
        assert_array_equal(actual.kcu, LineType.LINE_1D2D_GROUNDWATER)
    else:
        assert len(actual) == 0


@pytest.mark.parametrize(
    "n_groundwater_cells,expected", [(0, [-9999, -9999]), (3, [-9999, 5])]
)
def test_transfer_2d_node_to_groundwater(n_groundwater_cells, expected):
    lines_1d2d = Lines1D2D(id=range(1, 3), line=[[-9999, 6], [2, 9]])
    lines_1d2d.transfer_2d_node_to_groundwater(n_groundwater_cells)
    assert_array_equal(lines_1d2d.line[:, 0], expected)  # 2d side
    assert_array_equal(lines_1d2d.line[:, 1], [6, 9])  # 1d side


def test_assign_groundwater_exchange():
    nodes = Nodes(
        id=[1, 2, 3],
        content_type=[CN, CN, CH],
        content_pk=[3, 5, 3],
    )
    connection_nodes = ConnectionNodes(
        id=[3, 5],
        storage_area=[9.0, np.nan],
        exchange_thickness=[0.1, np.nan],
        hydraulic_conductivity_out=[3.0, np.nan],
        hydraulic_conductivity_in=[2.0, np.nan],
    )
    lines_1d2d = Lines1D2D(id=[1, 2, 3], line=[[-9999, 1], [-9999, 2], [-9999, 3]])
    lines_1d2d.assign_groundwater_exchange(nodes, connection_nodes)

    assert_almost_equal(lines_1d2d.hydraulic_resistance_in, [20.0, np.nan, np.nan])
    assert_almost_equal(lines_1d2d.hydraulic_resistance_out, [30.0, np.nan, np.nan])



"""The domain layer (main logic) of the project

This layer contains all the entities and logic surrounding those entities.

The domain layer does not depend on the interface or application layers.
"""

from .boundary_conditions import *  # NOQA
from .channels import *  # NOQA
from .connection_nodes import *  # NOQA
from .cross_section_definitions import *  # NOQA
from .cross_section_locations import *  # NOQA
from .dem_average_area import *  # NOQA
from .embedded import *  # NOQA
from .exchange_lines import *  # NOQA
from .grid import *  # NOQA
from .grid_refinement import *  # NOQA
from .lines_1d2d import *  # NOQA
from .obstacles import *  # NOQA
from .pipes import *  # NOQA
from .potential_breaches import *  # NOQA
from .quadtree import *  # NOQA
from .structures import *  # NOQA
from .windshielding import *  # NOQA
from .zero_d import *  # NOQA



import itertools

import numpy as np
import pytest
import shapely
from numpy.testing import assert_array_equal

from threedigrid_builder.exceptions import SchematisationError
from threedigrid_builder.grid import GridRefinements, QuadTree
from threedigrid_builder.grid.quadtree import reduce_refinement_levels


@pytest.fixture
def subgrid_meta():
    width = 20
    height = 16
    mask = np.ones((width, height), dtype=np.int8, order="F")
    mask[15:, :] = 0
    return {
        "pixel_size": 0.5,
        "width": width,
        "height": height,
        "bbox": (10.0, 10.0, 20.0, 18.0),
        "area_mask": mask,
    }


@pytest.fixture
def quadtree_no_refinement(subgrid_meta):
    return QuadTree(
        subgrid_meta=subgrid_meta,
        num_refine_levels=3,
        min_gridsize=1.0,
        use_2d_flow=True,
        refinements=None,
    )


@pytest.fixture
def quadtree_no_2d_flow(subgrid_meta):
    return QuadTree(
        subgrid_meta=subgrid_meta,
        num_refine_levels=3,
        min_gridsize=1.0,
        use_2d_flow=False,
        refinements=None,
    )


@pytest.fixture
def quadtree_line_refinement(subgrid_meta):
    refinement = GridRefinements(
        id=np.array([1]),
        refinement_level=np.array([1]),
        the_geom=shapely.linestrings([[[15.0, 17.5], [17.5, 17.5], [17.5, 14.5]]]),
    )
    return QuadTree(
        subgrid_meta=subgrid_meta,
        num_refine_levels=3,
        min_gridsize=1.0,
        use_2d_flow=True,
        refinements=refinement,
    )


@pytest.fixture
def quadtree_poly_refinement(subgrid_meta):
    refinement = GridRefinements(
        id=np.array([1]),
        refinement_level=np.array([2]),
        the_geom=np.array([shapely.box(12.4, 11.4, 13.9, 13.4)]),
    )
    return QuadTree(
        subgrid_meta=subgrid_meta,
        num_refine_levels=3,
        min_gridsize=1.0,
        use_2d_flow=True,
        refinements=refinement,
    )


def test_quadtree_no_refinement(quadtree_no_refinement):
    assert quadtree_no_refinement.kmax == 1
    assert np.size(quadtree_no_refinement.mmax) == 1
    assert quadtree_no_refinement.mmax[0] == 3
    assert quadtree_no_refinement.nmax[0] == 2
    assert quadtree_no_refinement.dx[0] == 4.0
    expected_lg = np.array(
        [
            [1, 1, -99],
            [1, 1, -99],
        ],
        dtype=np.int32,
    )
    assert_array_equal(quadtree_no_refinement.lg, expected_lg[::-1].T)


def test_quadtree_no_even_pixels(subgrid_meta):
    with pytest.raises(SchematisationError, match=r".*not contain an even number.*"):
        QuadTree(
            subgrid_meta=subgrid_meta,
            num_refine_levels=1,
            min_gridsize=1.5,
            use_2d_flow=True,
            refinements=None,
        )


def test_quadtree_line_refinement(quadtree_line_refinement):
    assert quadtree_line_refinement.kmax == 3
    assert np.size(quadtree_line_refinement.mmax) == 3
    assert quadtree_line_refinement.mmax[2] == 3
    assert quadtree_line_refinement.nmax[2] == 2
    assert quadtree_line_refinement.dx[0] == 1.0
    expected_lg = np.array(
        [
            [2, 2, 2, 2, 1, 1, 1, 1, -99, -99, -99, -99],
            [2, 2, 2, 2, 1, 1, 1, 1, -99, -99, -99, -99],
            [2, 2, 2, 2, 2, 2, 1, 1, -99, -99, -99, -99],
            [2, 2, 2, 2, 2, 2, 1, 1, -99, -99, -99, -99],
            [3, 3, 3, 3, 2, 2, 2, 2, -99, -99, -99, -99],
            [3, 3, 3, 3, 2, 2, 2, 2, -99, -99, -99, -99],
            [3, 3, 3, 3, 2, 2, 2, 2, -99, -99, -99, -99],
            [3, 3, 3, 3, 2, 2, 2, 2, -99, -99, -99, -99],
        ],
        dtype=np.int32,
    )
    assert_array_equal(quadtree_line_refinement.lg, expected_lg[::-1].T)


def test_quadtree_poly_refinement(quadtree_poly_refinement):
    assert quadtree_poly_refinement.kmax == 2
    assert np.size(quadtree_poly_refinement.mmax) == 2
    assert quadtree_poly_refinement.mmax[1] == 3
    assert quadtree_poly_refinement.nmax[1] == 2
    assert quadtree_poly_refinement.dx[0] == 2.0
    expected_lg = np.array(
        [
            [2, 2, 2, 2, -99, -99],
            [2, 2, 2, 2, -99, -99],
            [1, 1, 2, 2, -99, -99],
            [1, 1, 2, 2, -99, -99],
        ],
        dtype=np.int32,
    )
    assert_array_equal(quadtree_poly_refinement.lg, expected_lg[::-1].T)


def test_nodes_from_quadtree_line(quadtree_line_refinement, subgrid_meta):
    nodes, lines = quadtree_line_refinement.get_nodes_lines(
        subgrid_meta["area_mask"], itertools.count(start=0), itertools.count(start=0)
    )
    coordinates = np.array(
        [
            [12.0, 12.0],
            [11.0, 15.0],
            [11.0, 17.0],
            [13.0, 15.0],
            [13.0, 17.0],
            [15.0, 11.0],
            [15.0, 13.0],
            [15.0, 15.0],
            [17.0, 11.0],
            [17.0, 13.0],
            [14.5, 16.5],
            [14.5, 17.5],
            [15.5, 16.5],
            [15.5, 17.5],
            [16.5, 14.5],
            [16.5, 15.5],
            [16.5, 16.5],
            [16.5, 17.5],
            [17.5, 14.5],
            [17.5, 15.5],
            [17.5, 16.5],
            [17.5, 17.5],
        ]
    )

    pixel_coords_check = np.array(
        [[0, 0, 8, 8], [0, 12, 4, 16], [12, 0, 16, 4], [14, 14, 16, 16]]
    )
    assert_array_equal(nodes.coordinates, coordinates)
    assert_array_equal(nodes.pixel_coords[(0, 2, 8, 21), :], pixel_coords_check)


def test_lines_from_quadtree_line(quadtree_line_refinement, subgrid_meta):
    nodes, lines = quadtree_line_refinement.get_nodes_lines(
        subgrid_meta["area_mask"], itertools.count(start=0), itertools.count(start=0)
    )
    line = np.array(
        [
            [0, 5],
            [0, 6],
            [1, 3],
            [2, 4],
            [3, 7],
            [4, 10],
            [4, 11],
            [5, 8],
            [6, 9],
            [7, 14],
            [7, 15],
            [10, 12],
            [11, 13],
            [12, 16],
            [13, 17],
            [14, 18],
            [15, 19],
            [16, 20],
            [17, 21],
            [0, 1],
            [0, 3],
            [1, 2],
            [3, 4],
            [5, 6],
            [6, 7],
            [7, 10],
            [7, 12],
            [8, 9],
            [9, 14],
            [9, 18],
            [10, 11],
            [12, 13],
            [14, 15],
            [15, 16],
            [16, 17],
            [18, 19],
            [19, 20],
            [20, 21],
        ]
    )
    # fmt: off
    lik = np.array(
        [2, 2, 2, 2, 2, 1, 1, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2,
         2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=np.int32
    )
    lim = np.array(
        [3, 3, 1, 1, 2, 5, 5, 3, 3, 7, 7, 5, 5, 6, 6, 7, 7, 7, 7, 1, 2, 1, 2, 3,
         3, 5, 6, 4, 7, 8, 5, 6, 7, 7, 7, 8, 8, 8], dtype=np.int32
    )
    lin = np.array(
        [1, 2, 3, 4, 3, 7, 8, 1, 2, 5, 6, 7, 8, 7, 8, 5, 6, 7, 8, 3, 3, 3, 3, 1,
         2, 7, 7, 1, 5, 5, 7, 7, 5, 6, 7, 5, 6, 7], dtype=np.int32
    )
    # fmt: on

    cross_pix_coords = np.array(
        [
            [8, 0, 8, 4],
            [8, 4, 8, 8],
            [4, 8, 4, 12],
            [8, 12, 10, 12],
            [14, 8, 16, 8],
            [14, 10, 16, 10],
        ],
        dtype=np.int32,
    )

    assert_array_equal(lines.line, line)
    assert_array_equal(lines.lik, lik)
    assert_array_equal(lines.lim, lim)
    assert_array_equal(lines.lin, lin)
    assert_array_equal(
        lines.cross_pix_coords[(0, 1, 2, 25, 29, 35), :], cross_pix_coords
    )


def test_nodes_from_quadtree_poly(quadtree_poly_refinement, subgrid_meta):
    nodes, lines = quadtree_poly_refinement.get_nodes_lines(
        subgrid_meta["area_mask"], itertools.count(start=0), itertools.count(start=0)
    )
    expected_coordinates = [
        [12.0, 16.0],
        [16.0, 12.0],
        [16.0, 16.0],
        [11.0, 11.0],
        [11.0, 13.0],
        [13.0, 11.0],
        [13.0, 13.0],
    ]

    expected_pixel_coords = [
        [0, 8, 8, 16],
        [8, 0, 16, 8],
        [8, 8, 16, 16],
        [0, 0, 4, 4],
        [0, 4, 4, 8],
        [4, 0, 8, 4],
        [4, 4, 8, 8],
    ]
    assert_array_equal(nodes.coordinates, expected_coordinates)
    assert_array_equal(nodes.pixel_coords, expected_pixel_coords)


def test_lines_from_quadtree_poly(quadtree_poly_refinement, subgrid_meta):
    nodes, lines = quadtree_poly_refinement.get_nodes_lines(
        subgrid_meta["area_mask"], itertools.count(start=0), itertools.count(start=0)
    )
    line = [
        [0, 2],
        [5, 1],
        [6, 1],
        [3, 5],
        [4, 6],
        [4, 0],
        [6, 0],
        [1, 2],
        [3, 4],
        [5, 6],
    ]
    lik = [2, 1, 1, 1, 1, 1, 1, 2, 1, 1]
    lim = [1, 2, 2, 1, 1, 1, 2, 2, 1, 2]
    lin = [2, 1, 2, 1, 2, 2, 2, 1, 1, 1]

    cross_pix_coords = [
        [8, 8, 8, 16],
        [8, 0, 8, 4],
        [8, 4, 8, 8],
        [4, 0, 4, 4],
        [4, 4, 4, 8],
        [0, 8, 4, 8],
        [4, 8, 8, 8],
        [8, 8, 16, 8],
        [0, 4, 4, 4],
        [4, 4, 8, 4],
    ]
    assert_array_equal(lines.line, line)
    assert_array_equal(lines.lik, lik)
    assert_array_equal(lines.lim, lim)
    assert_array_equal(lines.lin, lin)
    assert_array_equal(lines.cross_pix_coords, cross_pix_coords)


def test_quarter_administration(quadtree_line_refinement, subgrid_meta):
    nodes, lines = quadtree_line_refinement.get_nodes_lines(
        subgrid_meta["area_mask"], itertools.count(start=0), itertools.count(start=0)
    )

    quarters = quadtree_line_refinement.get_quarters_admin(nodes, lines)

    # fmt: off
    neighbours = [
        [-9999, -9999],
        [5, -9999],
        [-9999, 1],
        [6, 3],
        [-9999, 0],
        [3, 0],
        [-9999, 2],
        [3, 2],
        [-9999, 1],
        [4, 1],
        [-9999, -9999],
        [4, -9999],
        [1, 0],
        [7, 0],
        [1, 4],
        [7, 4],
        [2, 3],
        [10, 3],
        [2, -9999],
        [11, -9999],
        [0, -9999],
        [8, -9999],
        [0, 6],
        [8, 6],
        [0, 5],
        [9, 5],
        [0, 7],
        [9, 7],
        [3, 6],
        [14, 6],
        [3, 10],
        [15, 12],
        [5, -9999],
        [-9999, -9999],
        [5, 9],
        [-9999, 9],
        [6, 8],
        [-9999, 8],
        [6, 14],
        [-9999, 18],
        [4, 7],
        [12, 7],
        [4, 11],
        [12, 11],
        [4, 10],
        [13, 10],
        [4, -9999],
        [13, -9999],
        [10, 7],
        [16, 7],
        [10, 13],
        [16, 13],
        [11, 12],
        [17, 12],
        [11, -9999],
        [17, -9999],
        [7, 9],
        [18, 9],
        [7, 15],
        [18, 15],
        [7, 14],
        [19, 14],
        [7, 16],
        [19, 16],
        [12, 15],
        [20, 15],
        [12, 17],
        [20, 17],
        [13, 16],
        [21, 16],
        [13, -9999],
        [21, -9999],
        [14, 9],
        [-9999, 9],
        [14, 19],
        [-9999, 19],
        [15, 18],
        [-9999, 18],
        [15, 20],
        [-9999, 20],
        [16, 19],
        [-9999, 19],
        [16, 21],
        [-9999, 21],
        [17, 20],
        [-9999, 20],
        [17, -9999],
        [-9999, -9999]
    ]

    lines = [
        [-9999, -9999],
        [0, -9999],
        [-9999, 19],
        [1, 20],
        [-9999, 19],
        [2, 19],
        [-9999, 21],
        [2, 21],
        [-9999, 21],
        [3, 21],
        [-9999, -9999],
        [3, -9999],
        [2, 20],
        [4, 20],
        [2, 22],
        [4, 22],
        [3, 22],
        [5, 22],
        [3, -9999],
        [6, -9999],
        [0, -9999],
        [7, -9999],
        [0, 23],
        [7, 23],
        [1, 23],
        [8, 23],
        [1, 24],
        [8, 24],
        [4, 24],
        [9, 24],
        [4, 25],
        [10, 26],
        [7, -9999],
        [-9999, -9999],
        [7, 27],
        [-9999, 27],
        [8, 27],
        [-9999, 27],
        [8, 28],
        [-9999, 29],
        [5, 25],
        [11, 25],
        [5, 30],
        [11, 30],
        [6, 30],
        [12, 30],
        [6, -9999],
        [12, -9999],
        [11, 26],
        [13, 26],
        [11, 31],
        [13, 31],
        [12, 31],
        [14, 31],
        [12, -9999],
        [14, -9999],
        [9, 28],
        [15, 28],
        [9, 32],
        [15, 32],
        [10, 32],
        [16, 32],
        [10, 33],
        [16, 33],
        [13, 33],
        [17, 33],
        [13, 34],
        [17, 34],
        [14, 34],
        [18, 34],
        [14, -9999],
        [18, -9999],
        [15, 29],
        [-9999, 29],
        [15, 35],
        [-9999, 35],
        [16, 35],
        [-9999, 35],
        [16, 36],
        [-9999, 36],
        [17, 36],
        [-9999, 36],
        [17, 37],
        [-9999, 37],
        [18, 37],
        [-9999, 37],
        [18, -9999],
        [-9999, -9999]
    ]
    # fmt: on
    assert_array_equal(quarters.line, lines)
    assert_array_equal(quarters.neighbour_node, neighbours)


def test_no_2d_flow(quadtree_no_2d_flow, subgrid_meta):
    nodes, lines = quadtree_no_2d_flow.get_nodes_lines(
        subgrid_meta["area_mask"],
        itertools.count(start=0),
        itertools.count(start=0),
    )

    assert len(lines) == 0
    assert len(nodes) == 4


@pytest.mark.parametrize(
    "kmax,refinement_level,new_kmax,new_refinement_level",
    [
        (3, 1, 3, 1),
        (3, 2, 2, 1),
        (3, 3, 1, 1),
    ],
)
def test_reduce_refinement_levels(
    kmax, refinement_level, new_kmax, new_refinement_level
):
    refinements = GridRefinements(
        id=[1],
        refinement_level=[refinement_level],
        the_geom=shapely.linestrings([[[15.0, 17.5], [17.5, 17.5], [17.5, 14.5]]]),
    )
    actual = reduce_refinement_levels(refinements, kmax)

    assert actual == new_kmax
    assert (refinements.refinement_level == new_refinement_level).all()


@pytest.mark.parametrize("refinements", [None, GridRefinements(id=[])])
def test_reduce_refinement_levels_no_refinements(refinements):
    assert reduce_refinement_levels(refinements, 3) == 1


def test_quadtree_no_active_pixels(subgrid_meta):
    subgrid_meta["area_mask"][:] = 0
    with pytest.raises(SchematisationError):
        QuadTree(
            subgrid_meta=subgrid_meta,
            num_refine_levels=3,
            min_gridsize=1.0,
            use_2d_flow=True,
            refinements=None,
        )



import itertools

import numpy as np
import shapely

from threedigrid_builder.base import Array, Lines
from threedigrid_builder.constants import CalculationType, ContentType, FrictionType
from threedigrid_builder.grid import linear

__all__ = ["Culverts", "Orifices", "Weirs"]


class Culvert:  # NL: duiker
    id: int
    code: str
    the_geom: shapely.Geometry
    dist_calc_points: float
    connection_node_start_id: int
    connection_node_end_id: int
    calculation_type: CalculationType
    cross_section_definition_id: int
    invert_level_start_point: float
    invert_level_end_point: float
    discharge_coefficient_negative: float
    discharge_coefficient_positive: float
    friction_type: FrictionType
    friction_value: float
    zoom_category: int
    display_name: str


class Culverts(Array[Culvert], linear.BaseLinear):
    content_type = ContentType.TYPE_V2_CULVERT

    def is_closed(self, content_pk):
        """Whether objects are 'closed' or 'open water'.

        This is relevant for 1D-2D connections.
        """
        return np.full(len(content_pk), True, dtype=bool)

    def get_1d2d_exchange_levels(self, content_pk, s1d, connection_nodes, **kwargs):
        """Compute the exchange level (dpumax) for 1D-2D flowlines.

        For pipes and culverts 1D2D exchange levels are interpololated between
        connection node drain levels.

        Args:
            content_pk (array of int): object ids for which to compute levels
            s1d (array of int): distance on the objects where to compute levels
            connection_nodes (ConnectionNodes): for the drain_levels

        Returns:
            exchange levels a.k.a. dpumax (array of float)
        """
        return self.compute_drain_level(
            ids=content_pk,
            s=s1d,
            connection_nodes=connection_nodes,
        )


class WeirOrifice:  # NL: stuw / doorlaat
    id: int
    code: str
    connection_node_start_id: int
    connection_node_end_id: int
    crest_level: float
    crest_type: CalculationType
    cross_section_definition_id: int
    discharge_coefficient_negative: float
    discharge_coefficient_positive: float
    friction_type: FrictionType
    friction_value: float
    zoom_category: int
    display_name: str
    sewerage: int


class WeirOrifices(Array[WeirOrifice]):
    content_type = None  # subclasses need to define this

    def get_lines(self, connection_nodes, line_id_counter, connection_node_offset=0):
        """Convert weirs (or orifices) into lines

        Args:
            connection_nodes (ConnectionNodes): used to map ids to indices
            line_id_counter (iterable): an iterable yielding integers
            connection_node_offset (int): offset to give connection node
              indices in the returned lines.line. Default 0.

        Returns:
            Lines with data in the following columns:
            - id: counter generated from line_id_counter
            - line: 2 node ids per line
            - content_type: the content type of the structure (weir or orifice)
            - content_pk: the id of the structure from which this line originates
            - kcu: the crest_type of the structure
            - dpumax: the crest_level of the structure
            - cross_id1 & cross_id2: the id of the cross section definition
            - cross_weight: 1.0 (which means that cross_id2 should be ignored)
            - frict_type1 & frict_type2: the friction type (both are equal)
            - frict_value1 & frict_value2: the friction value (both are equal)
            - invert_level_start_point: the crest_level of the structure
            - invert_level_end_point: the crest_level of the structure
            - discharge_coefficient_positive and _negative: taken from the structure
            - display_name: user defined label for display purposes
        """
        # map connection node IDs to node indices
        line = np.empty((len(self), 2), dtype=np.int32, order="F")
        line[:, 0] = connection_nodes.id_to_index(self.connection_node_start_id)
        line[:, 1] = connection_nodes.id_to_index(self.connection_node_end_id)
        line += connection_node_offset
        return Lines(
            id=itertools.islice(line_id_counter, len(self)),
            line=line,
            content_type=self.content_type,
            content_pk=self.id,
            kcu=self.crest_type,  # implicitly converts CalculationType -> LineType
            dpumax=self.crest_level,
            cross_id1=self.cross_section_definition_id,
            cross_id2=self.cross_section_definition_id,
            cross_weight=1.0,
            frict_type1=self.friction_type,
            frict_value1=self.friction_value,
            frict_type2=self.friction_type,
            frict_value2=self.friction_value,
            invert_level_start_point=self.crest_level,
            invert_level_end_point=self.crest_level,
            discharge_coefficient_positive=self.discharge_coefficient_positive,
            discharge_coefficient_negative=self.discharge_coefficient_negative,
            display_name=self.display_name,
            connection_node_start_id=self.connection_node_start_id,
            connection_node_end_id=self.connection_node_end_id,
            crest_level=self.crest_level,
            crest_type=self.crest_type,
            sewerage=self.sewerage,
        )


class Orifices(WeirOrifices):
    content_type = ContentType.TYPE_V2_ORIFICE


class Weirs(WeirOrifices):
    content_type = ContentType.TYPE_V2_WEIR



import numpy as np

from threedigrid_builder.base import RasterInterface

try:
    from osgeo import gdal

    gdal.UseExceptions()
except ImportError:
    gdal = None

__all__ = ["GDALInterface"]

GT_TOLERANCE = 7


class GDALInterface(RasterInterface):
    def __init__(self, *args, **kwargs):
        if gdal is None:
            raise ImportError("Cannot use GDALInterface if GDAL is not available.")
        super().__init__(*args, **kwargs)
        if self.model_area_path is not None:
            raise NotImplementedError("Model areas are not available with GDAL")

    def __enter__(self):
        self._dataset = gdal.Open(self.path.as_posix(), gdal.GA_ReadOnly)
        c, a, b, f, d, e = self._dataset.GetGeoTransform()
        self.set_transform((a, b, c, d, e, f))
        self.set_crs(self._dataset.GetProjection())
        return self

    def __exit__(self, *args, **kwargs):
        del self._dataset

    def read(self):
        width, height, bbox, mask = self._create_area_arr_from_dem()

        return {
            "pixel_size": self.pixel_size,
            "width": width,
            "height": height,
            "bbox": bbox,
            "area_mask": np.flipud(mask).T.astype(dtype=np.int8, copy=True, order="F"),
        }

    def _create_area_arr_from_dem(self):
        xpixel, _, xmin, _, ypixel, ymax = self.transform
        width, height = self._dataset.RasterXSize, self._dataset.RasterYSize
        xmax = xmin + width * xpixel
        ymin = ymax + height * ypixel

        if xmax < xmin:
            xmin, xmax = xmax, xmin
        if ymax < ymin:
            ymin, ymax = ymax, ymin

        band = self._dataset.GetRasterBand(1)
        size_j, size_i = band.GetBlockSize()
        nodata = band.GetNoDataValue()
        mask = np.zeros((height, width), dtype=np.int8)
        n_blocks_j = ((width - 1) // size_j) + 1
        n_blocks_i = ((height - 1) // size_i) + 1

        for i in range(n_blocks_i):
            for j in range(n_blocks_j):
                i1 = i * size_i
                j1 = j * size_j
                i2 = min(i1 + size_i, height)
                j2 = min(j1 + size_j, width)
                data = band.ReadAsArray(
                    xoff=j1, yoff=i1, win_xsize=j2 - j1, win_ysize=i2 - i1
                )
                _mask = np.isfinite(data)
                if nodata is not None and np.isfinite(nodata):
                    _mask &= data != nodata
                mask[i1:i2, j1:j2] = _mask
        return width, height, (xmin, ymin, xmax, ymax), mask



"""The application layer a.k.a. use cases of the project

Use cases orchestrate the flow of data to and from the domain entities.

This layer depends on the interfaces as well as on the domain layer.
"""

import itertools
import logging
from pathlib import Path
from typing import Callable, Optional

from threedigrid_builder.base import GridSettings
from threedigrid_builder.base.surfaces import Surfaces
from threedigrid_builder.constants import InflowType
from threedigrid_builder.exceptions import SchematisationError
from threedigrid_builder.grid import Grid, QuadTree
from threedigrid_builder.interface import (
    DictOut,
    GDALInterface,
    GeopackageOut,
    GridAdminOut,
    SQLite,
)

__all__ = ["make_grid", "make_gridadmin"]

logger = logging.getLogger(__name__)


def _default_progress_callback(progress: float, message: str):
    logger.info("Progress: %d, Message: %s", progress * 100, message)


def _make_gridadmin(
    sqlite_path,
    dem_path=None,
    meta=None,
    progress_callback=None,
    upgrade=False,
    convert_to_geopackage=False,
):
    """Compute interpolated channel nodes"""
    progress_callback(0.0, "Reading input schematisation...")
    db = SQLite(
        sqlite_path, upgrade=upgrade, convert_to_geopackage=convert_to_geopackage
    )

    node_id_counter = itertools.count()
    line_id_counter = itertools.count()
    embedded_node_id_counter = itertools.count()

    settings = db.get_settings()
    grid = Grid.from_meta(**settings, **(meta or {}))
    grid_settings: GridSettings = settings["grid_settings"]

    if grid_settings.use_2d:
        progress_callback(0.1, "Reading subgrid input...")
        if not dem_path:
            raise SchematisationError("DEM file expected")
        with GDALInterface(dem_path) as raster:
            subgrid_meta = raster.read()

        # Patch CRS with that of the DEM (so: user-supplied EPSG is ignored)
        grid.set_crs(raster.crs)
        db.epsg_code = grid.meta.epsg_code

        refinements = db.get_grid_refinements()
        progress_callback(0.7, "Constructing 2D computational grid...")
        quadtree = QuadTree(
            subgrid_meta,
            grid_settings.kmax,
            grid_settings.grid_space,
            grid_settings.use_2d_flow,
            refinements,
        )
        grid += Grid.from_quadtree(
            quadtree=quadtree,
            area_mask=subgrid_meta["area_mask"],
            node_id_counter=node_id_counter,
            line_id_counter=line_id_counter,
        )
        grid.set_quarter_administration(quadtree)

        if grid.meta.has_groundwater:
            grid.add_groundwater(
                grid.meta.has_groundwater_flow, node_id_counter, line_id_counter
            )

        grid.set_obstacles(db.get_obstacles())
        grid.set_boundary_conditions_2d(
            db.get_boundary_conditions_2d(),
            quadtree,
            node_id_counter,
            line_id_counter,
        )

        dem_average_areas = db.get_dem_average_areas()
        grid.set_dem_averaged_cells(dem_average_areas)

    connection_nodes = db.get_connection_nodes()
    if grid_settings.use_1d_flow and len(connection_nodes) > 0:
        progress_callback(0.8, "Constructing 1D computational grid...")
        cn_grid = Grid.from_connection_nodes(
            connection_nodes=connection_nodes,
            node_id_counter=node_id_counter,
        )
        connection_node_first_id = cn_grid.nodes.id[0] if len(cn_grid.nodes) > 0 else 0
        grid += cn_grid

        channels = db.get_channels()
        potential_breaches = db.get_potential_breaches()
        breach_points = potential_breaches.project_on_channels(channels).merge()

        channel_grid = Grid.from_linear_objects(
            connection_nodes=connection_nodes,
            objects=channels,
            fixed_nodes=breach_points,
            cell_tree=grid.cell_tree if grid_settings.use_2d else None,
            global_dist_calc_points=grid_settings.dist_calc_points,
            embedded_cutoff_threshold=grid_settings.embedded_cutoff_threshold,
            node_id_counter=node_id_counter,
            embedded_node_id_counter=embedded_node_id_counter,
            line_id_counter=line_id_counter,
            connection_node_offset=connection_node_first_id,
        )

        locations = db.get_cross_section_locations()
        locations.apply_to_lines(channel_grid.lines, channels)
        windshieldings = db.get_windshieldings()
        windshieldings.apply_to_lines(channel_grid.lines)
        grid += channel_grid

        pipes = db.get_pipes()
        grid += Grid.from_linear_objects(
            connection_nodes=connection_nodes,
            objects=pipes,
            fixed_nodes=None,
            cell_tree=grid.cell_tree if grid_settings.use_2d else None,
            global_dist_calc_points=grid_settings.dist_calc_points,
            embedded_cutoff_threshold=grid_settings.embedded_cutoff_threshold,
            node_id_counter=node_id_counter,
            embedded_node_id_counter=embedded_node_id_counter,
            line_id_counter=line_id_counter,
            connection_node_offset=connection_node_first_id,
        )

        culverts = db.get_culverts()
        grid += Grid.from_linear_objects(
            connection_nodes=connection_nodes,
            objects=culverts,
            fixed_nodes=None,
            cell_tree=grid.cell_tree if grid_settings.use_2d else None,
            global_dist_calc_points=grid_settings.dist_calc_points,
            embedded_cutoff_threshold=grid_settings.embedded_cutoff_threshold,
            node_id_counter=node_id_counter,
            embedded_node_id_counter=embedded_node_id_counter,
            line_id_counter=line_id_counter,
            connection_node_offset=connection_node_first_id,
        )

        weirs = db.get_weirs()
        orifices = db.get_orifices()
        grid += grid.from_structures(
            connection_nodes=connection_nodes,
            weirs=weirs,
            orifices=orifices,
            line_id_counter=line_id_counter,
            connection_node_offset=connection_node_first_id,
        )

        grid.set_calculation_types()
        grid.set_bottom_levels()
        grid.set_breach_ids(breach_points)
        grid.set_initial_waterlevels(
            connection_nodes=connection_nodes,
            channels=channels,
            pipes=pipes,
            culverts=culverts,
        )
        grid.set_boundary_conditions_1d(db.get_boundary_conditions_1d())
        grid.set_cross_sections(db.get_cross_section_definitions())
        grid.set_pumps(db.get_pumps())

    if grid.nodes.has_1d and grid.nodes.has_2d:
        progress_callback(0.9, "Connecting 1D and 2D domains...")
        grid.embed_nodes(embedded_node_id_counter)

        grid.add_1d2d_lines(
            exchange_lines=db.get_exchange_lines(),
            connection_nodes=connection_nodes,
            channels=channels,
            pipes=pipes,
            locations=locations,
            culverts=culverts,
            potential_breaches=potential_breaches,
            line_id_counter=line_id_counter,
        )

        if grid.meta.has_groundwater:
            channels.apply_has_groundwater_exchange(grid.nodes, grid.lines)
            pipes.apply_has_groundwater_exchange(grid.nodes, grid.lines)
            grid.add_1d2d_groundwater_lines(line_id_counter, connection_nodes)

    if grid_settings.use_0d_inflow in (
        InflowType.IMPERVIOUS_SURFACE.value,
        InflowType.SURFACE.value,
    ):
        # process zero-d information, either using the 'v2_surfaces'
        # or 'v2_impervious_surfaces' table.
        progress_callback(0.95, "Processing 0D domain...")
        surfaces: Surfaces = db.get_surfaces()
        grid.add_0d(surfaces)

    grid.finalize()
    return grid


def make_gridadmin(
    sqlite_path: Path,
    dem_path: Path,
    out_path: Optional[Path] = None,
    meta: dict = None,
    progress_callback: Optional[Callable[[float, str], None]] = None,
    upgrade: bool = False,
    convert_to_geopackage: bool = False,
):
    """Create a Grid instance from sqlite and DEM paths

    The SQLite is expected to be validated already with threedi-modelchecker.

    Args:
        sqlite_path: The path to the input schematisation (SQLite) file
        dem_path: The path of the input DEM file (GeoTIFF)
        out_path: The path of the (to be created) output file. Allowed extensions
            are: .h5 (HDF5) and .gpkg (Geopackage). If not supplied, this function will
            return an in-memory representation of the Grid.
        meta: an optional dict with the following (optional) keys: model_slug (str),
            revision_hash (str), revision_nr (int), threedi_version (str)
        progress_callback: an optional function that updates the progress. The function
            should take an float in the range 0-1 and a message string.
        upgrade: whether to upgrade the sqlite (inplace) before processing

    Raises:
        threedigrid_builder.SchematisationError: if there is something wrong with
            the input schematisation (SQLite) file.
    """
    if isinstance(sqlite_path, str):
        sqlite_path = Path(sqlite_path)
    if isinstance(dem_path, str):
        dem_path = Path(dem_path)
    if out_path is None:
        Writer = DictOut
    else:
        if isinstance(out_path, str):
            out_path = Path(out_path)
        extension = out_path.suffix.lower()
        if extension == ".h5":
            Writer = GridAdminOut
        elif extension == ".gpkg":
            Writer = GeopackageOut
        else:
            raise ValueError(f"Unsupported output format '{extension}'")

    if not Writer.available():
        raise ImportError(
            "This output format is not available because dependencies are missing."
        )
    if progress_callback is None:
        progress_callback = _default_progress_callback

    grid = _make_gridadmin(
        sqlite_path,
        dem_path,
        meta=meta,
        progress_callback=progress_callback,
        upgrade=upgrade,
        convert_to_geopackage=convert_to_geopackage,
    )

    progress_callback(0.99, "Writing gridadmin...")
    with Writer(out_path) as writer:
        return writer.write(grid)


# Legacy
make_grid = make_gridadmin
